r"""
Uses the building detector to validate a construction event.

Given a site model:

* search for the K highest quality images at the start of the sequences.

* search for the K highest quality images at the end of the sequence.

* run the building detector on all of the chosen images.

* test if the site boundary intersects detections in the start images.

* test if the site boundary intersects detections in the end images.


.. code::

    StartTest | EndTest | Result
    ----------+---------+-------
        T     |    T    | Reject
    ----------+---------+-------
        T     |    F    | Reject
    ----------+---------+-------
        F     |    T    | Accept
    ----------+---------+-------
        F     |    F    | Reject
    ----------+---------+-------
        ?     |    T    | Accept
    ----------+---------+-------
        T     |    ?    | Reject
    ----------+---------+-------
        ?     |    F    | Reject
    ----------+---------+-------
        F     |    ?    | Accept
    ----------+---------+-------


Dataflow:

    * BAS outputs a region model with candidate site summaries

    * We should be given a kwcoco path that indexes all of the data we could
      look at. In MLOPs this will be a region cropped kwcoco path that indexes
      existant images on disk. In smartflow this will be a virtual kwcoco file
      that requires network access.
"""
#!/usr/bin/env python3
import scriptconfig as scfg
import ubelt as ub


class BuildingValidatorConfig(scfg.DataConfig):
    input_kwcoco = scfg.Value(None, help='input')

    input_region = scfg.Value(None)
    input_sites = scfg.Value(None)

    output_region_fpath = scfg.Value(None)
    output_sites_dpath = scfg.Value(None)
    output_site_manifest_fpath = scfg.Value(None)

    box_isect_threshold = scfg.Value(0.1, help='This fraction of the a detected building box must intersect the proposed polygon')
    box_score_threshold = scfg.Value(0.01, help='The detected building boxes must have a score higher than this')
    start_max_score = scfg.Value(1.0, help='The max building score needed in the start sequence')
    end_min_score = scfg.Value(0.1, help='The min building score needed in the end sequence')


IGNORE_CLASS_NAMES = ub.codeblock(
    '''
    Fixed-wing Aircraft
    Small Aircraft
    Cargo Plane
    Helicopter
    Passenger Vehicle
    Small Car
    Bus
    Pickup Truck
    Utility Truck
    Truck
    Cargo Truck
    Truck w/Box
    Truck Tractor
    Trailer
    Truck w/Flatbed
    Truck w/Liquid
    Crane Truck
    Railway Vehicle
    Passenger Car
    Cargo Car
    Flat Car
    Tank car
    Locomotive
    Maritime Vessel
    Motorboat
    Sailboat
    Tugboat
    Barge
    Fishing Vessel
    Ferry
    Yacht
    Container Ship
    Oil Tanker
    Engineering Vehicle
    Tower crane
    Container Crane
    Reach Stacker
    Straddle Carrier
    Mobile Crane
    Dump Truck
    Haul Truck
    Scraper/Tractor
    Front loader/Bulldozer
    Excavator
    Cement Mixer
    Ground Grader
    ''').split('\n')

DONT_IGNORE_CLASSNAMES = ub.codeblock(
    '''
    Aircraft Hangar
    Helipad
    Storage Tank
    Hut/Tent
    Shipping container lot
    Shipping Container
    Pylon
    Tower
    Shed
    Building
    Damaged Building
    Facility
    Construction Site
    Vehicle Lot
    ''').split('\n')


def main(cmdline=1, **kwargs):
    """
    Ignore:
        from watch.tasks.dino_detector.building_validator import *  # NOQA
        NODE_DPATH = '/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_fd298dba'
        from watch.utils.partial_format import fsubtemplate
        coco_fpath = fsubtemplate('$NODE_DPATH/pred_boxes.kwcoco.zip')
        from watch.utils.util_path import coerce_patterned_paths
        site_summary = ub.Path(fsubtemplate('$NODE_DPATH/.pred/valicrop/*/.pred/bas_poly/*/site_summaries_manifest.json'))
        site_summary = coerce_patterned_paths(site_summary)[0]

        output_region_fpath = ub.Path(fsubtemplate('$NODE_DPATH/filtered_site_summaries.geojson'))

        kwargs = {
            'input_kwcoco': coco_fpath,
            'input_region': site_summary,
            'output_region_fpath': output_region_fpath,
        }
        cmdline = 0

    Ignore:
        >>> import watch
        >>> dvc_data_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        >>> coco_fpath = dvc_data_dpath / 'Drop6/imgonly-KR_R001.kwcoco.json'
        >>> region_fpath = dvc_data_dpath / 'annotations/drop6_hard_v1/region_models/KR_R001.geojson'

    Example:
        >>> # xdoctest: +SKIP
        >>> cmdline = 0
        >>> kwargs = dict()
        >>> main(cmdline=cmdline, **kwargs)
    """
    config = BuildingValidatorConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    rich.print('config = ' + ub.urepr(config, nl=1))

    from watch.geoannots import geomodels
    from watch.utils import util_gis
    import kwcoco
    from watch.cli import reproject_annotations
    from watch.utils import util_time
    from kwcoco.util import util_json
    from watch.utils import process_context
    import os
    import safer
    import json

    # Args will be serailized in kwcoco, so make sure it can be coerced to json
    jsonified_config = util_json.ensure_json_serializable(config.asdict())
    walker = ub.IndexableWalker(jsonified_config)
    for problem in util_json.find_json_unserializable(jsonified_config):
        bad_data = problem['data']
        walker[problem['loc']] = str(bad_data)
    filter_output = {
        'type': 'tracking_result',
        'info': [],
        'files': [],
    }
    # Track process info
    proc_context = process_context.ProcessContext(
        name='watch.tasks.dino_detector.building_validator', type='process',
        config=jsonified_config,
        track_emissions=False,
    )
    proc_context.start()
    filter_output['info'].append(proc_context.obj)

    input_coco = kwcoco.CocoDataset(config.input_kwcoco)
    region_model = geomodels.RegionModel.coerce(config.input_region)
    output_region_fpath = ub.Path(config.output_region_fpath)

    input_site_fpaths = util_gis.coerce_geojson_paths(config.input_sites)
    # set(input_coco.annots().lookup('track_id', None))

    site_id_to_summary = {}
    for summary in region_model.site_summaries():
        assert summary.site_id not in site_id_to_summary
        site_id_to_summary[summary.site_id] = summary
    ##

    output_kwcoco = reproject_annotations.main(
        cmdline=0, src=input_coco.copy(),
        dst='return',
        region_models=config.input_region,
        status_to_catname={'system_confirmed': 'positive'},
        role='pred_poly',
        validate_checks=False,
        clear_existing=False,
    )

    site_to_accept = {}
    for video_id in ub.ProgIter(output_kwcoco.videos(), desc='validate sites'):
        video = output_kwcoco.index.videos[video_id]
        video_name = video['name']  # the vide name should be the site id
        summary = site_id_to_summary[video_name]
        site_id = summary.site_id
        video_images = output_kwcoco.images(video_id=video_id)

        start_date = summary.start_date
        end_date = summary.end_date

        # Get the starting and ending observations
        start_images = []
        end_images = []
        for coco_img in video_images.coco_images:
            img_time = util_time.coerce_datetime(coco_img['date_captured'])
            dist_start = abs(img_time - start_date)
            dist_end = abs(img_time - end_date)
            if dist_end < dist_start:
                end_images.append(coco_img)
            else:
                start_images.append(coco_img)

        try:
            start_features = []
            for coco_img in start_images:
                feat = building_in_image_features(coco_img, site_id, config)
                start_features.append(feat)

            end_features = []
            for coco_img in end_images:
                feat = building_in_image_features(coco_img, site_id, config)
                end_features.append(feat)

            if len(start_features) and len(end_features):
                max_start_score = max(f['max_score'] for f in start_features)
                max_end_score = max(f['max_score'] for f in end_features)
                accept = (
                    max_start_score <= config.start_max_score and
                    max_end_score >= config.end_min_score
                )
            else:
                # Unobservable case, automatically accept
                accept = True
        except CouldNotValidate:
            accept = True
        site_to_accept[site_id] = accept

    accept_sites = [s for s, f in site_to_accept.items() if f]
    print(f'Filter to {len(accept_sites)} / {len(site_id_to_summary)} sites')
    new_summaries = ub.udict(site_id_to_summary).subdict(accept_sites).values()

    site_to_site_fpath = ub.udict({
        p.stem: p for p in input_site_fpaths
    })

    # assert set(site_to_site_fpath) == set(site_id_to_summary)
    sites_with_paths = set(site_to_site_fpath)
    sites_with_summary = set(site_id_to_summary)
    if sites_with_paths != sites_with_summary:
        print('sites_with_paths = {}'.format(ub.urepr(sites_with_paths, nl=1)))
        print('sites_with_summary = {}'.format(ub.urepr(sites_with_summary, nl=1)))
        raise AssertionError(
            f'sites with paths {len(sites_with_paths)} are not the same as '
            f'sites with summaries {len(sites_with_summary)}')

    # Copy the filtered site models over to the output directory
    keep_site_fpaths = site_to_site_fpath.subdict(accept_sites)
    output_sites_dpath = ub.Path(config.output_sites_dpath)
    output_sites_dpath.ensuredir()
    for old_fpath in keep_site_fpaths.values():
        new_fpath = output_sites_dpath / old_fpath.name
        old_fpath.copy(new_fpath, overwrite=True)

    new_region_model = geomodels.RegionModel.from_features(
        [region_model.header] + list(new_summaries))

    output_region_fpath.parent.ensuredir()
    with safer.open(output_region_fpath, 'w', temp_file=not ub.WIN32) as file:
        json.dump(new_region_model, file, indent=4)

    proc_context.stop()

    if config.output_site_manifest_fpath is not None:
        filter_output['files'] = [os.fspath(output_region_fpath)]
        print(f'Write filtered site result to {config.output_site_manifest_fpath}')
        with safer.open(config.output_site_manifest_fpath, 'w', temp_file=not ub.WIN32) as file:
            json.dump(filter_output, file, indent=4)


def building_in_image_features(coco_img, site_id, config):
    import numpy as np
    import kwimage
    import geopandas as gpd
    import warnings
    annots = coco_img.annots()
    flags = np.array([r == 'pred_poly' for r in annots.lookup('role', None)])
    box_annots = annots.compress(~flags)
    poly_annots = annots.compress(flags)
    is_main_poly = [t == site_id for t in poly_annots.lookup('track_id')]
    main_poly_annots = poly_annots.compress(is_main_poly)
    if len(main_poly_annots) > 1:
        warnings.warn('FIXME: Len of "main-poly-annots" is not 1.')
    elif len(main_poly_annots) == 0:
        raise CouldNotValidate('We dont expect to be here')

    flags = [cname not in IGNORE_CLASS_NAMES for cname in box_annots.category_names]
    box_annots = box_annots.compress(flags)

    iooa_thresh = config.box_isect_threshold
    score_thresh = config.box_score_threshold

    box_gdf = gpd.GeoDataFrame({
        'geometry': box_annots.boxes.to_shapely(),
        'class': box_annots.category_names,
        'score': box_annots.lookup('score'),
    })
    box_gdf = box_gdf[box_gdf.score > score_thresh]

    proposal_gdf = gpd.GeoDataFrame({
        'geometry': [
            kwimage.MultiPolygon.coerce(p).to_shapely()
            for p in main_poly_annots.lookup('segmentation')
        ],
    })
    proposal_geom = proposal_gdf.iloc[0]['geometry']

    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'invalid value encountered in intersection')
        intersections = box_gdf.intersection(proposal_geom)
        bad_flags = (intersections.isnull() | intersections.is_empty)
        isect_flags = (~bad_flags) & (intersections.is_valid)
        feat = {
            'max_iou': 0,
            'max_iooa': 0,
            'max_score': 0,
            'num_cands': 0,
        }
        if isect_flags.any():
            cand_boxes = box_gdf[isect_flags]
            cand_isects = intersections[isect_flags]
            cand_iooa1 = cand_isects.area / cand_boxes.area
            ioaa_flags = cand_iooa1 > iooa_thresh
            if ioaa_flags.any():
                cand_boxes = cand_boxes[ioaa_flags]

                cand_unions = cand_boxes.union(proposal_geom)
                cand_ious = cand_isects.area / cand_unions.area

                feat['max_iou'] = cand_ious.max()
                feat['max_iooa'] = cand_iooa1.max()
                feat['max_score'] = cand_boxes['score'].max()
                feat['num_cands'] = len(cand_boxes)
        return feat


class CouldNotValidate(Exception):
    ...


if __name__ == '__main__':
    """

    CommandLine:
        python ~/code/watch/watch/tasks/dino_detector/building_validator.py
        python -m watch.tasks.dino_detector.building_validator
    """
    main()

"""

Ignore:
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
    geowatch schedule --params="
        matrix:
            bas_pxl.package_fpath:
                - $DVC_EXPT_DPATH/models/fusion/Drop6-MeanYear10GSD/packages/Drop6_TCombo1Year_BAS_10GSD_split6_V42_cont2/Drop6_TCombo1Year_BAS_10GSD_split6_V42_cont2_epoch3_step941.pt
            bas_pxl.test_dataset:
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-KR_R002.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-BR_R002.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-CH_R001.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-NZ_R001.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-KR_R001.kwcoco.zip
                - $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-AE_R001.kwcoco.zip
            bas_pxl.chip_overlap: 0.3
            bas_pxl.chip_dims:
                - auto
            bas_pxl.time_span:
                - auto
            bas_pxl.time_sampling:
                # - auto
                # - soft5
                - soft4
            bas_poly.thresh:
                - 0.33
                #- 0.38
                #- 0.4
            bas_poly.inner_window_size:
                - 1y
                #- null
            bas_poly.inner_agg_fn:
                - mean
            bas_poly.norm_ord:
                #- 1
                - inf
            bas_poly.polygon_simplify_tolerance:
                - 1
            bas_poly.agg_fn:
                - probs
            bas_poly.resolution:
                - 10GSD
            bas_poly.moving_window_size:
                - null
                #- 1
            bas_poly.poly_merge_method:
                - 'v2'
                #- 'v1'
            bas_poly.min_area_square_meters:
                - 7200
            bas_poly.max_area_square_meters:
                - 8000000
            bas_poly.boundary_region: $DVC_DATA_DPATH/annotations/drop6/region_models
            bas_poly_eval.true_site_dpath: $DVC_DATA_DPATH/annotations/drop6/site_models
            bas_poly_eval.true_region_dpath: $DVC_DATA_DPATH/annotations/drop6/region_models
            bas_pxl.enabled: 1
            bas_pxl_eval.enabled: 1
            bas_poly.enabled: 1
            bas_poly_eval.enabled: 1
            bas_poly_viz.enabled: 0
            valicrop.enabled: 1
            valicrop.minimum_size: "256x256@2GSD"
            valicrop.num_start_frames: 3
            valicrop.num_end_frames: 3
            valicrop.context_factor: 1.5
            sv_dino_boxes.enabled: 1
            sv_dino_boxes.package_fpath: $DVC_EXPT_DPATH/models/kitware/xview_dino.pt
            sv_dino_boxes.window_dims:
                - 256
                - 512
                - 768
                - 1024
                # - 1536
            sv_dino_boxes.window_overlap:
                - "0.5"
            sv_dino_boxes.fixed_resolution:
                - "1GSD"
                - "2GSD"
                - "2.5GSD"
                - "3GSD"
                - "3.3GSD"
            sv_dino_filter.box_isect_threshold:
                - 0.1
            sv_dino_filter.box_score_threshold:
                - 0.01
            sv_dino_filter.start_max_score:
                - 1.0
                - 0.8
                # - 0.5
            sv_dino_filter.end_min_score:
                - 0.0
                # - 0.05
                - 0.1
                - 0.15
                - 0.2
                - 0.25
                - 0.3
                # - 0.4
                - 0.5
        submatrices:
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-KR_R001.kwcoco.zip
              valicrop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-KR_R001.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-KR_R002.kwcoco.zip
              valicrop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-KR_R002.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-BR_R002.kwcoco.zip
              valicrop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-BR_R002.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-CH_R001.kwcoco.zip
              valicrop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-CH_R001.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-NZ_R001.kwcoco.zip
              valicrop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-NZ_R001.kwcoco.json
            - bas_pxl.test_dataset: $DVC_DATA_DPATH/Drop6-MeanYear10GSD-V2/imganns-AE_R001.kwcoco.zip
              valicrop.crop_src_fpath: $DVC_DATA_DPATH/Drop6/imgonly-AE_R001.kwcoco.json
        " \
        --root_dpath="$DVC_EXPT_DPATH/_mlops_eval10_baseline" \
        --devices="0," --tmux_workers=8 \
        --backend=tmux --queue_name "_mlops_eval10_baseline" \
        --pipeline=bas_building_vali --skip_existing=1 \
        --run=1

DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)
gwmlops aggregate \
    --pipeline=bas_building_vali \
    --target \
        "$DVC_EXPT_DPATH/_mlops_eval10_baseline" \
    --resource_report=0 \
    --rois='[KR_R002,KR_R001,BR_R002,CH_R001,NZ_R001,AE_R001]' \
    --stdout_report="
        top_k: 5
        per_group: 1
        macro_analysis: 0
        analyze: 0
        reference_region: final
        shorten: 1
    " --eval_nodes="[sv_poly_eval]" \
    --plot_params="
        enabled: True
    "

    --rois='[NZ_R001,KR_R001]' \



        print_models: True

geowatch align \
    --src "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_data_dvc-ssd/Drop6/imgonly-KR_R001.kwcoco.json" \
    --dst "/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/valicrop/valicrop_id_2e8c8dc3/valicrop.kwcoco.zip" \
    --regions="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/bas_poly/bas_poly_id_dc32b2a6/site_summaries_manifest.json" \
    --site_summary=True \
    --verbose="1" \
    --workers="32" \
    --aux_workers="4" \
    --debug_valid_regions="False" \
    --visualize="False" \
    --keep="img" \
    --geo_preprop="auto"  \
    --minimum_size="256x256@2GSD" \
    --num_start_frames="3" \
    --num_end_frames="3" \
    --context_factor="1.5" \
    --include_sensors="WV" \
    --force_nodata="-9999" \
    --rpc_align_method="orthorectify" \
    --target_gsd="2" \
    --force_min_gsd="2" \
    --convexify_regions="True"

geowatch visualize /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/valicrop/valicrop_id_2e8c8dc3/valicrop.kwcoco.zip --smart

python ~/code/watch/dev/wip/grid_sitevali_crops.py /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/valicrop/valicrop_id_2e8c8dc3/_viz_*


python -m watch.tasks.dino_detector.predict \
    --package_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/models/kitware/xview_dino.pt" \
    --coco_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/valicrop/valicrop_id_2e8c8dc3/valicrop.kwcoco.zip" \
    --out_coco_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_61b8c2c7/pred_boxes.kwcoco.zip" \
    --device="0" \
    --data_workers="2" \
    --fixed_resolution="1.0GSD" \
    --window_dims="2048" \
    --batch_size="1"

geowatch visualize /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_61b8c2c7/pred_and_truth.kwcoco.zip \
    --resolution=2GSD \
    --smart \
    --ann_score_thresh=0.3 \
    --viz_dpath /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_61b8c2c7/_vizme

python ~/code/watch/dev/wip/grid_sitevali_crops.py --sub=_anns \
    /home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_61b8c2c7/_vizme


NODE_DPATH=/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_fd298dba
DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)

geowatch reproject \
        --src "$NODE_DPATH/pred_boxes.kwcoco.zip" \
        --dst "$NODE_DPATH/pred_boxes_with_polys.kwcoco.zip" \
        --region_models "$NODE_DPATH/.pred/valicrop/*/.pred/bas_poly/*/site_summaries_manifest.json" \
        --status_to_catname="{system_confirmed: positive}" \
        --role=pred_poly \
        --validate_checks=False \
        --clear_existing=False

geowatch reproject \
        --src "$NODE_DPATH/pred_boxes_with_polys.kwcoco.zip" \
        --dst "$NODE_DPATH/pred_and_truth.kwcoco.zip" \
        --region_models="$DVC_DATA_DPATH/annotations/drop6/region_models/*.geojson" \
        --site_models="$DVC_DATA_DPATH/annotations/drop6/site_models/*.geojson" \
        --status_to_catname="{system_confirmed: positive}" \
        --role=truth \
        --clear_existing=False

gw visualize --smart 1 \
    --ann_score_thresh 0.5 \
    --draw_labels False \
    --alpha 0.5 \
    --src $NODE_DPATH/pred_and_truth.kwcoco.zip \
    --viz_dpath $NODE_DPATH/_vizme \

python ~/code/watch/dev/wip/grid_sitevali_crops.py \
    --sub=_anns \
    $NODE_DPATH/_vizme

python -m watch.tasks.dino_detector.building_validator \
    --input_kwcoco "$NODE_DPATH/pred_boxes.kwcoco.zip" \
    --input_region $NODE_DPATH/.pred/valicrop/*/.pred/bas_poly/*/site_summaries_manifest.json \
    --output_region_fpath "$NODE_DPATH/filtered_summaries.json" \
    --box_isect_threshold 0.1 \
    --box_score_threshold 0.1 \
    --start_max_score 0.0 \
    --end_min_score 0.3

python -m watch.tasks.dino_detector.building_validator \
    --input_kwcoco="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/buildings/buildings_id_663bd461/pred_boxes.kwcoco.zip" \
    --input_region="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/bas_poly/bas_poly_id_f8061df6/site_summaries_manifest.json" \
    --input_sites="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/bas_poly/bas_poly_id_f8061df6/sites_manifest.json" \
    --output_region_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/building_validate/building_validate_id_60a56eed/out_region.geojson" \
    --output_sites_dpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/building_validate/building_validate_id_60a56eed/out_sites" \
    --output_site_manifest_fpath="/home/joncrall/remote/toothbrush/data/dvc-repos/smart_expt_dvc/_mlops_eval10_baseline/pred/flat/building_validate/building_validate_id_60a56eed/out_site_manifest.json" \
    --box_isect_threshold="0.1" \
    --box_score_threshold="0.1" \
    --start_max_score="0.1" \
    --end_min_score="0.3"


"""



import kwarray
import kwimage
import numpy as np
import torch
import ubelt as ub
from functools import partial
from netharn.data.batch_samplers import PatchedBatchSampler
from netharn.data.data_containers import ItemContainer
from netharn.data.data_containers import container_collate
from netharn.data.batch_samplers import PatchedRandomSampler
from netharn.data.batch_samplers import SubsetSampler
import random


class SequenceDataset(torch.utils.data.Dataset):
    """
    Example:
        >>> from watch.tasks.rutgers_material_seg.datasets.iarpa_contrastive_dataset import *  # NOQA
        >>> import ndsampler
        >>> import itertools as it
        >>> sampler = ndsampler.CocoSampler.demo('vidshapes8', image_size=(64, 64))
        >>> channels = 'r|g|b'
        >>> window_dims = (2, 128, 128)
        >>> self = SequenceDataset(sampler, window_dims=window_dims, training=False, channels=channels)
        >>> index_iter = it.count()
        >>> index = next(index_iter)
        >>> item = self[index]
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> import einops
        >>> kwplot.autompl()
        >>> frames = item['inputs']['im'].data
        >>> frame_masks = item['label']['class_masks'].data
        >>> frames_ = einops.rearrange(frames, 'c t h w -> t c h w').numpy()
        >>> frames_ = kwimage.normalize_intensity(frames_)
        >>> frames_ = np.nan_to_num(frames_)
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(frames_))
        >>> for frame, mask in zip(frames_, frame_masks):
        >>>     kwplot.imshow(frame, pnum=pnum_())
        >>>     heatmap = kwimage.Heatmap(class_idx=mask, classes=self.sampler.classes)
        >>>     heatmap.draw(with_alpha=0.3)

    Example:
        >>> from watch.tasks.rutgers_material_seg.datasets.iarpa_contrastive_dataset import *  # NOQA
        >>> import ndsampler
        >>> import itertools as it
        >>> sampler = ndsampler.CocoSampler.demo('vidshapes8-msi', image_size=(64, 64))
        >>> channels = 'B1|B8|B11'
        >>> window_dims = (2, 128, 128)
        >>> self = SequenceDataset(sampler, window_dims=window_dims, training=False, inference_only=True, channels=channels)
        >>> index_iter = it.count()
        >>> index = next(index_iter)
        >>> item = self[index]
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> import einops
        >>> kwplot.autompl()
        >>> frames = item['inputs']['im'].data
        >>> assert 'label' not in item
        >>> frames_ = einops.rearrange(frames, 'c t h w -> t c h w').numpy()
        >>> frames_ = kwimage.normalize_intensity(frames_)
        >>> frames_ = np.nan_to_num(frames_)
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(frames_))
        >>> for frame in frames_:
        >>>     kwplot.imshow(frame, pnum=pnum_())
    """

    def __init__(self, sampler, window_dims, input_dims=None, channels=None,
                 rng=None, training=True, window_overlap=0.0,
                 inference_only=False):
        super().__init__()
        if input_dims is None:
            input_dims = window_dims[-2:]
        # print(sampler)
        self.training = training
        self.rng = kwarray.ensure_rng(rng)
        self.sampler = sampler
        self.window_dims = window_dims
        self.input_dims = input_dims
        self.classes = self.sampler.classes
        self.channels = channels
        self.inference_only = inference_only
        # Build a simple space-time-grid
        sample_grid_spec = {
            'task': 'video_detection',
            'window_dims': window_dims,
            'window_overlap': window_overlap,
        }
        # print(sample_grid_spec)
        self.sample_grid = sampler.new_sample_grid(**sample_grid_spec)
        self.training = training

        if self.training:
            self.chosen_samples = self.sample_grid['positives']
        else:
            # In inference we need to load all positive and negative regions
            self.chosen_samples = (
                self.sample_grid['positives'] +
                self.sample_grid['negatives']
            )
            # Reorganize samples so they iterate through images / videos in
            # order
            grouped = ub.group_items(
                self.chosen_samples,
                lambda x: tuple(
                    [x['vidid']] + [gid for gid in x['gids']]
                )
            )
            grouped = ub.sorted_keys(grouped)
            self.chosen_samples = list(ub.flatten(grouped.values()))

    def __len__(self):
        return len(self.chosen_samples)

    def __getitem__(self, index):

        sampler = self.sampler

        tr = self.chosen_samples[index]
        if self.channels:
            tr['channels'] = self.channels

        if not self.inference_only:
            with_annots = 'segmentation'
        else:
            with_annots = False

        # NOTE: Setting nodata=float ensures this returns data in a floating
        # format and also ensures any values that need to be masked are filled
        # with nans.
        sample: dict = sampler.load_sample(
            tr, with_annots=with_annots, nodata='float',
            padkw={'mode': 'constant', 'constant_values': np.nan})

        if self.training:
            # Only need to get a "negative" in training
            negative_index = random.randint(0, self.__len__() - 2)
            tr_negative = self.chosen_samples[negative_index]
            if self.channels:
                tr_negative['channels'] = self.channels
            negative_sample: dict = sampler.load_sample(
                tr_negative, with_annots=with_annots, nodata='float',
                padkw={'mode': 'constant', 'constant_values': np.nan})

        # Access the sampled image and annotation data
        raw_frame_list = sample['im']

        if not self.inference_only:
            raw_det_list = sample['annots']['frame_dets']
        else:
            raw_det_list = [None] * len(raw_frame_list)

        # Break data down on a per-frame basis so we can apply image-based
        # augmentations.
        frame_ims = []
        frame_masks = []
        for raw_frame, raw_dets in zip(raw_frame_list, raw_det_list):
            frame = raw_frame.astype(np.float32)
            input_dsize = self.input_dims[-2:][::-1]
            # Resize the sampled window to the target space for the network
            frame, info = kwimage.imresize(frame, dsize=input_dsize,
                                           interpolation='linear',
                                           return_info=True)

            # ensure channel dim is not squeezed
            frame = kwarray.atleast_nd(frame, 3)
            frame_ims.append(frame)

            if not self.inference_only:
                # Remember to apply any transform to the dets as well
                dets: kwimage.Detections = raw_dets
                dets = dets.scale(info['scale'])
                # print(info)
                dets = dets.translate(info['offset'])
                # print(frame.shape[0:2])
                frame_mask = np.full(frame.shape[0:2], dtype=np.int32, fill_value=-1)
                ann_polys = dets.data['segmentations'].to_polygon_list()
                # print(ann_polys)
                ann_aids = dets.data['aids']
                ann_cids = dets.data['cids']

                for poly, aid, cid in zip(ann_polys, ann_aids, ann_cids):
                    cidx = self.classes.id_to_idx[cid]
                    poly.fill(frame_mask, value=cidx)
                frame_masks.append(frame_mask)

        # Perpare data for torch
        frame_data = np.concatenate([f[None, ...] for f in frame_ims], axis=0)
        cthw_im = frame_data.transpose(3, 0, 1, 2)
        # print(f"image min:{cthw_im.min()}, max:{cthw_im.max()}")
        inputs = {
            'im': ItemContainer(torch.from_numpy(cthw_im).contiguous(), stack=True),
        }

        if not self.inference_only:
            class_masks = np.concatenate([m[None, ...] for m in frame_masks], axis=0)
            label = {
                'class_masks': ItemContainer(
                    torch.from_numpy(class_masks).contiguous(), stack=False, cpu_only=True),
            }

        if self.training:
            negative_raw_frame_list = negative_sample['im']
            negative_raw_det_list = negative_sample['annots']['frame_dets']
            negative_frame_ims = []
            # negative_frame_masks = []
            for negative_raw_frame, negative_raw_dets in zip(negative_raw_frame_list, negative_raw_det_list):
                frame = negative_raw_frame.astype(np.float32)
                # dets = negative_raw_dets
                input_dsize = self.input_dims[-2:][::-1]
                # Resize the sampled window to the target space for the network
                frame, info = kwimage.imresize(frame, dsize=input_dsize,
                                               interpolation='linear',
                                               return_info=True)
                # Remember to apply any transform to the dets as well
                # dets = dets.scale(info['scale'])
                # dets = dets.translate(info['offset'])
                # frame_mask = np.full(frame.shape[0:2], dtype=np.int32, fill_value=-1)
                # ann_polys = dets.data['segmentations'].to_polygon_list()
                # ann_aids = dets.data['aids']
                # ann_cids = dets.data['cids']
                # for poly, aid, cid in zip(ann_polys, ann_aids, ann_cids):
                #     cidx = self.classes.id_to_idx[cid]
                #     poly.fill(frame_mask, value=cidx)

                # ensure channel dim is not squeezed
                frame = kwarray.atleast_nd(frame, 3)

                # fig = plt.figure()
                # ax1 = fig.add_subplot(1,2,1)
                # ax2 = fig.add_subplot(1,2,2)
                # ax1.imshow(frame[:,:,:3])
                # ax2.imshow(frame_mask)
                # plt.show()

                negative_frame_ims.append(frame)
                # negative_frame_masks.append(frame_mask)

            # Prepare data for torch
            negative_frame_data = np.concatenate([f[None, ...] for f in negative_frame_ims], axis=0)
            negative_cthw_im = negative_frame_data.transpose(3, 0, 1, 2)
            # UNUSED? FIXME?
            # negative_class_masks = np.concatenate([m[None, ...] for m in negative_frame_masks], axis=0)  # NOQA
            inputs['negative_im'] = ItemContainer(torch.from_numpy(negative_cthw_im).contiguous(), stack=True)

        item = {
            'inputs': inputs,
            'tr': ItemContainer(sample['tr'], stack=False),
        }
        if not self.inference_only:
            item['label'] = label

        return item

    def make_loader(self, batch_size=16, num_workers=0, shuffle=False,
                    pin_memory=True, drop_last=True, multiscale=False,
                    num_batches='auto', xpu=None):
        """
        Create a loader for this dataset with custom sampling logic and
        container collation.
        """
        if len(self) == 0:
            raise ValueError('must have some data')

        # The case where where replacement is not allowed
        if num_batches == 'auto':
            num_samples = None
        else:
            num_samples = num_batches * batch_size

        if shuffle:
            item_sampler = PatchedRandomSampler(self, num_samples=num_samples)
        else:
            if num_samples is None:
                item_sampler = torch.utils.data.sampler.SequentialSampler(self)
            else:
                stats_idxs = (np.arange(num_samples) % len(self))
                item_sampler = SubsetSampler(stats_idxs)

        if num_samples is not None:
            # If num_batches is too big, error
            if num_samples > len(self):
                raise IndexError(
                    'num_batches={} and batch_size={} causes '
                    'num_samples={} to be greater than the number '
                    'of data items {}. Try setting num_batches=auto?'.format(
                        num_batches, batch_size, num_samples, len(self)))

        batch_sampler = PatchedBatchSampler(
            item_sampler, batch_size=batch_size, drop_last=drop_last,
            num_batches=num_batches)

        if xpu is None:
            num_devices = 1
        else:
            num_devices = len(xpu.devices)

        collate_fn = partial(container_collate, num_devices=num_devices)

        loader = torch.utils.data.DataLoader(
            self, batch_sampler=batch_sampler,
            collate_fn=collate_fn, num_workers=num_workers,
            pin_memory=pin_memory, worker_init_fn=worker_init_fn)
        return loader


def worker_init_fn(worker_id):
    worker_info = torch.utils.data.get_worker_info()
    self = worker_info.dataset

    if hasattr(self.sampler.dset, 'connect'):
        # Reconnect to the backend if we are using SQL
        self.sampler.dset.connect(readonly=True)



# flake8: noqa

import torch
import torch.nn as nn
import math
import torch.utils.model_zoo as model_zoo
from torch.nn import functional as F
from watch.tasks.rutgers_material_seg.models.sg import StochasticGate
from watch.tasks.rutgers_material_seg.models.gci import GCI

__all__ = ['ResNet', 'resnet50', 'resnet101', 'resnet152']


model_urls = {
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}


def rescale_as(x, y, mode="bilinear", align_corners=True):
    h, w = y.size()[2:]
    x = F.interpolate(x, size=[h, w], mode=mode, align_corners=align_corners)
    return x


class Conv2d(nn.Conv2d):

    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True):
        super(Conv2d, self).__init__(in_channels, out_channels, kernel_size, stride,
                                     padding, dilation, groups, bias)

    def forward(self, x):
        # return super(Conv2d, self).forward(x)
        weight = self.weight
        weight_mean = weight.mean(dim=(1, 2, 3), keepdim=True)
        weight = weight - weight_mean
        std = weight.view(weight.size(0), -1).std(dim=1).view(-1, 1, 1, 1) + 1e-5
        weight = weight / std.expand_as(weight)
        return F.conv2d(x, weight, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)


class ASPP(nn.Module):

    def __init__(self, C, depth, num_classes, conv=nn.Conv2d,
                 norm=nn.BatchNorm2d, momentum=0.0003, mult=1):
        super(ASPP, self).__init__()
        self._C = C
        self._depth = depth
        self._num_classes = num_classes

        self.global_pooling = nn.AdaptiveAvgPool2d(1)
        self.relu = nn.ReLU(inplace=True)
        self.aspp1 = conv(C, depth, kernel_size=1, stride=1, bias=False)
        self.aspp2 = conv(C, depth, kernel_size=3, stride=1,
                          dilation=int(6 * mult), padding=int(6 * mult),
                          bias=False)
        self.aspp3 = conv(C, depth, kernel_size=3, stride=1,
                          dilation=int(12 * mult), padding=int(12 * mult),
                          bias=False)
        self.aspp4 = conv(C, depth, kernel_size=3, stride=1,
                          dilation=int(18 * mult), padding=int(18 * mult),
                          bias=False)
        self.aspp5 = conv(C, depth, kernel_size=1, stride=1, bias=False)
        self.aspp1_bn = norm(depth, momentum)
        self.aspp2_bn = norm(depth, momentum)
        self.aspp3_bn = norm(depth, momentum)
        self.aspp4_bn = norm(depth, momentum)
        self.aspp5_bn = norm(depth, momentum)
        self.conv2 = conv(depth * 5, depth, kernel_size=1, stride=1,
                          bias=False)
        self.bn2 = norm(depth, momentum)
        self.conv3 = nn.Conv2d(depth, num_classes, kernel_size=1, stride=1)

    def forward(self, x):
        x1 = self.aspp1(x)
        x1 = self.aspp1_bn(x1)
        x1 = self.relu(x1)
        x2 = self.aspp2(x)
        x2 = self.aspp2_bn(x2)
        x2 = self.relu(x2)
        x3 = self.aspp3(x)
        x3 = self.aspp3_bn(x3)
        x3 = self.relu(x3)
        x4 = self.aspp4(x)
        x4 = self.aspp4_bn(x4)
        x4 = self.relu(x4)
        x5 = self.global_pooling(x)
        x5 = self.aspp5(x5)
        x5 = self.aspp5_bn(x5)
        x5 = self.relu(x5)
        x5 = nn.Upsample((x.shape[2], x.shape[3]), mode='bilinear',
                         align_corners=True)(x5)
        x = torch.cat((x1, x2, x3, x4, x5), 1)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.conv3(x)

        return x


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, dilation=1, conv=None, norm=None):
        super(Bottleneck, self).__init__()
        self.conv1 = conv(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = norm(planes)
        self.conv2 = conv(planes, planes, kernel_size=3, stride=stride,
                          dilation=dilation, padding=dilation, bias=False)
        self.bn2 = norm(planes)
        self.conv3 = conv(planes, planes * self.expansion, kernel_size=1,
                          bias=False)
        self.bn3 = norm(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.double_conv = nn.Sequential(
            Conv2d(in_channels, out_channels, kernel_size=3, stride=1, padding=1),
            nn.GroupNorm(32, out_channels),
            nn.LeakyReLU(0.2),
            nn.Dropout(0.5),
            Conv2d(out_channels, out_channels, kernel_size=3, stride=1, padding=1),
            nn.GroupNorm(32, out_channels),
            nn.LeakyReLU(0.2),
        )

    def forward(self, x):
        return self.double_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, bilinear=True):
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of
        # channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
        else:
            self.up = nn.ConvTranspose2d(in_channels // 2, in_channels // 2, kernel_size=2, stride=2)

        self.conv = DoubleConv(in_channels, out_channels)
        # Given transposed=1, weight of size [48, 48, 2, 2], 48 -> 32+64//2, instead,
        # expected input[4, 64, 128, 128] to have 48 channels, but got 64
        # channels instead

    def forward(self, x1, x2):
        # print(x1.shape)
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2])

        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class ResNet(nn.Module):

    def __init__(self, block, layers, num_classes, num_groups=None,
                 weight_std=True, beta=False, num_channels=3 , feats=None, out_dim=None):
        self.inplanes = 64
        feats = [256, 512, 1024, 2048, 256]

        def _norm(planes, momentum=0.05):
            if num_groups is None:
                return nn.BatchNorm2d(planes, momentum=momentum)
            else:
                return nn.GroupNorm(num_groups, planes)
        self.norm = _norm
        self.conv = Conv2d if weight_std else nn.Conv2d

        super(ResNet, self).__init__()
        if not beta:
            self.conv1_ = self.conv(num_channels, 64, kernel_size=3, stride=1,
                                   padding=1, bias=False)
        else:
            self.conv1_ = nn.Sequential(
                self.conv(num_channels, 64, 3, stride=1, padding=1, bias=False),
                # self.conv(64, 64, 3, stride=1, padding=1, bias=False),
                # self.conv(64, 64, 3, stride=1, padding=1, bias=False)
                )
        self.bn1 = self.norm(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=1, padding=1)
        self.layer1 = self._make_layer(block, feats[0] // block.expansion, layers[0])
        self.layer2 = self._make_layer(block, feats[1] // block.expansion, layers[1], stride=2)
        self.layer3 = self._make_layer(block, feats[2] // block.expansion, layers[2], stride=2)
        self.layer4 = self._make_layer(block, feats[3] // block.expansion, layers[3], stride=2, dilation=2)
        # self.aspp = ASPP(512 * block.expansion, 256, num_classes, conv=self.conv, norm=self.norm)
        self.aspp = ASPP(feats[3], 256, 256, conv=self.conv, norm=self.norm)

        for m in self.modules():
            if isinstance(m, self.conv):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d) or isinstance(m, nn.GroupNorm):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

        # self.shallow_mask = GCI()
        # self.from_scratch_layers += self.shallow_mask.from_scratch_layers

        # Stochastic Gate
        # self.sg = StochasticGate()
        self.up1 = Up(feats[4] + feats[3], feats[3], bilinear=True)
        self.up2 = Up(feats[2] + feats[3], feats[2], bilinear=True)
        self.up3 = Up(feats[1] + feats[2], feats[1], bilinear=True)
        # self.up4 = Up(feats[0] + feats[1], feats[0], bilinear=True)

        # self.fc8_skip = nn.Sequential(Conv2d(256, 48, 1, bias=False),
        #                               # nn.BatchNorm2d(48, track_running_stats = False),
        #                               nn.GroupNorm(24, 48),
        #                               nn.LeakyReLU(0.2))
        # self.fc8_x = nn.Sequential(Conv2d(560, 256, kernel_size=3, stride=1, padding=1, bias=False),
        #                            #    nn.BatchNorm2d(256, track_running_stats = False),
        #                            nn.GroupNorm(32, 256),
        #                            nn.LeakyReLU(0.2))

        # decoder
        # self.last_conv = nn.Sequential(Conv2d(256, 256, kernel_size=3, stride=1, padding=1, bias=False),
        #                                # nn.BatchNorm2d(256, track_running_stats = False),
        #                                nn.GroupNorm(32, 256),
        #                                nn.LeakyReLU(0.2),
        #                                nn.Dropout(0.5),
        #                                Conv2d(256, 256, kernel_size=3, stride=1, padding=1, bias=False),
        #                                # nn.BatchNorm2d(256, track_running_stats = False),
        #                                nn.GroupNorm(32, 256),
        #                                nn.LeakyReLU(0.2),
        #                                nn.Dropout(0.1),
        #                                nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True),
        #                                nn.LeakyReLU(0.2),
        #                                nn.Conv2d(256, 256, kernel_size=1, stride=1),
        #                                nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True),
        #                                nn.LeakyReLU(0.2),
        #                                nn.Conv2d(256, num_classes, kernel_size=1, stride=1))
        self.last_conv = Conv2d(feats[0], num_classes, kernel_size=1, stride=1, bias=False)

    def _make_layer(self, block, planes, blocks, stride=1, dilation=1):
        downsample = None
        if stride != 1 or dilation != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                self.conv(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, dilation=max(1, dilation / 2), bias=False),
                self.norm(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample,
                            dilation=max(1, dilation / 2), conv=self.conv,
                            norm=self.norm))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation,
                                conv=self.conv, norm=self.norm))

        return nn.Sequential(*layers)

    def forward(self, x):
        # size = (x.shape[2], x.shape[3])
        x1 = self.conv1_(x)
        x1 = self.bn1(x1)
        x1 = self.relu(x1)
        x1 = self.maxpool(x1)
        x1 = self.layer1(x1)
        # conv3 = x1
        # print(conv3.shape)
        x2 = self.layer2(x1)
        x3 = self.layer3(x2)
        x4 = self.layer4(x3)
        x_feats = self.aspp(x4)

        # x2_x = self.fc8_skip(conv3)
        # x_up = rescale_as(x_feats, x2_x)
        # x = self.fc8_x(torch.cat([x_up, x2_x], 1))
        x = self.up1(x_feats, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        # x = self.up4(x, x1)
        # 3.2 deep feature context for shallow features
        # x2 = self.shallow_mask(conv3, x)
        # 3.3 stochastically merging the masks
        # x = self.sg(x, x2, alpha_rate=0.3)
        # x = self.last_conv(x)
        # x = nn.Upsample(size, mode='bilinear', align_corners=True)(x)
        return x, x4


def resnet50(pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    # if pretrained:
    #     model.load_state_dict(model_zoo.load_url(model_urls['resnet50']))
    if pretrained:
        model_dict = model.state_dict()
        pretrained_path = "/home/native/projects/data/smart_watch/models/experiments_onera/tasks_experiments_onera_2021-10-07-10:23/experiments_epoch_34_loss_2151.7745061910377_valmIoU_0.5357620684181676_time_2021-10-09-07:21:41.pth"
        pretrained_dict = torch.load(pretrained_path)
        # pretrained_dict = model_zoo.load_url(model_urls['resnet50'])
        overlap_dict = {k: v for k, v in pretrained_dict.items()
                        if k in model_dict}
        model_dict.update(overlap_dict)
        model.load_state_dict(model_dict)
    return model


def resnet34(pretrained=False, **kwargs):
    """Constructs a ResNet-34 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    # if pretrained:
    #     model.load_state_dict(model_zoo.load_url(model_urls['resnet50']))
    if pretrained:
        model_dict = model.state_dict()
        pretrained_path = "/home/native/projects/data/smart_watch/models/experiments_onera/tasks_experiments_onera_2021-10-07-10:23/experiments_epoch_34_loss_2151.7745061910377_valmIoU_0.5357620684181676_time_2021-10-09-07:21:41.pth"
        pretrained_dict = torch.load(pretrained_path)['model']
        print(pretrained_dict.keys())
        # print(pretrained_dict.values())
        print(model_dict.keys())
        # pretrained_dict = model_zoo.load_url(model_urls['resnet50'])
        overlap_dict = {k[7:]: v for k, v in pretrained_dict.items()
                        if k[7:] in model_dict}
        print(f"loaded {len(overlap_dict)} layers")
        model_dict.update(overlap_dict)
        model.load_state_dict(model_dict)
    exit()
    return model


def resnet18(pretrained=False, **kwargs):
    """Constructs a ResNet-50 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [2, 2, 2, 2], **kwargs)
    # if pretrained:
    #     model.load_state_dict(model_zoo.load_url(model_urls['resnet50']))
    if pretrained:
        model_dict = model.state_dict()
        pretrained_dict = model_zoo.load_url(model_urls['resnet50'])
        overlap_dict = {k: v for k, v in pretrained_dict.items()
                        if k in model_dict}
        model_dict.update(overlap_dict)
        model.load_state_dict(model_dict)
    return model


def resnet101(pretrained=False, num_groups=None, weight_std=False, **kwargs):
    """Constructs a ResNet-101 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 4, 23, 3],
                   num_groups=num_groups, weight_std=weight_std, **kwargs)
    if pretrained:
        model_dict = model.state_dict()
        if num_groups and weight_std:
            pretrained_path = '/home/native/projects/data/smart_watch/models/R-101-GN-WS.pth.tar'
            # pretrained_path = "/home/native/projects/data/smart_watch/models/experiments_onera/tasks_experiments_onera_2021-10-07-10:23/experiments_epoch_34_loss_2151.7745061910377_valmIoU_0.5357620684181676_time_2021-10-09-07:21:41.pth"
            pretrained_dict = torch.load(pretrained_path)
            # print(pretrained_dict['conv1'])
            overlap_dict = {k[7:]: v for k, v in pretrained_dict.items()
                            if k[7:] in model_dict}
            # assert len(overlap_dict) == 312, len(overlap_dict)
            print(f"loaded {len(overlap_dict)} layers")
        elif not num_groups and not weight_std:
            pretrained_dict = model_zoo.load_url(model_urls['resnet101'])
            overlap_dict = {k: v for k, v in pretrained_dict.items()
                            if k in model_dict}
        else:
            raise ValueError('Currently only support BN or GN+WS')
        model_dict.update(overlap_dict)
        model.load_state_dict(model_dict, strict=False)
    return model


def resnet152(pretrained=False, **kwargs):
    """Constructs a ResNet-152 model.

    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    """
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet152']))
    return model



#!/usr/bin/env python3
"""
CommandLine:
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware='auto')
    python -m watch.cli.geojson_site_stats.py \
        --site_models="$DVC_DATA_DPATH/annotations/drop6/site_models/*"
"""
import scriptconfig as scfg
import ubelt as ub


class GeojsonSiteStatConfig(scfg.DataConfig):
    """
    Compute statistics about geojson sites
    """
    site_models = scfg.Value(None, help='site model coercable', nargs='+')

    viz_dpath = None


def main(cmdline=1, **kwargs):
    """
    Ignore:
        from watch.cli.geojson_site_stats import *  # NOQA
        import watch
        data_dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        kwargs = {
            'site_models': data_dvc_dpath / 'annotations/drop6/site_models/*',
        }
        cmdline = 0

        kwargs['site_models'] = '/data/joncrall/dvc-repos/smart_expt_dvc/smartflow_evals/kit_pre_eval_8_20230131/BR_R001/sc_out_site_models'
    """
    config = GeojsonSiteStatConfig.cli(cmdline=cmdline, data=kwargs, strict=True)
    import rich
    rich.print('config = ' + ub.repr2(config))
    import pandas as pd
    import numpy as np
    from watch.utils import util_gis
    from watch.utils import util_time
    import matplotlib.dates as mdates

    site_infos = list(util_gis.coerce_geojson_datas(config['site_models']))

    viz_dpath = config['viz_dpath']
    if viz_dpath is None:
        viz_dpath = '_viz_sitestat'
    viz_dpath = ub.Path(viz_dpath).ensuredir()

    obs_stats_accum = []
    site_stat_accum = []
    for site_info in site_infos:

        data_crs84 = site_info['data']
        data_utm = util_gis.project_gdf_to_local_utm(data_crs84)
        type_to_datas = dict(list(data_utm.groupby('type')))
        obs_df = type_to_datas.pop('observation', [])
        site_df = type_to_datas.pop('site', [])
        assert len(type_to_datas) == 0
        site_row = site_df.iloc[0]

        site_df = geopandas_shape_stats(site_df)

        obs_df = geopandas_shape_stats(obs_df)
        obs_df['region_id'] = site_row.region_id
        metric_keys = ['rt_area', 'major_obox_ratio', 'obox_major', 'obox_minor', 'hull_rt_area']

        keep_keys = ['status', 'region_id', 'site_id', 'current_phase', 'observation_date'] + metric_keys
        obs_subdf = obs_df[keep_keys].copy()

        valid_obs_rows = obs_subdf[~pd.isnull(obs_df['observation_date'])]
        valid_obs_rows = valid_obs_rows.sort_values('observation_date')
        obs_unixtimes = [util_time.coerce_datetime(d).timestamp() for d in valid_obs_rows['observation_date']]
        duration = np.diff(obs_unixtimes)
        obs_subdf.loc[valid_obs_rows.index[0:-1], 'duration'] = duration

        multiphase_to_duration = obs_subdf.groupby('current_phase')['duration'].sum()
        phase_to_duration = ub.ddict(lambda: 0)
        for mk, v in multiphase_to_duration.items():
            for k in mk.split(','):
                phase_to_duration['duration.' + k.strip()] += v

        keep_keys = ['status', 'region_id', 'site_id', 'start_date', 'end_date'] + metric_keys
        site_subdf = site_df[keep_keys].copy()
        durr_df = pd.DataFrame(ub.udict(phase_to_duration).map_values(lambda x: [x]))
        site_subdf = pd.concat([site_subdf, durr_df], axis=1)

        # pred_info_fpath = site_info['fpath'].parent.parent / 'site_tracks_manifest.json'
        # if pred_info_fpath.exists():
        #     import json
        #     info_section = smart_result_parser.parse_json_header(pred_info_fpath)
        #     track_kw = json.loads(info_section[-1]['properties']['args']['track_kwargs'])
        #     stats_df.loc[:, 'thresh'] = track_kw['thresh']
        obs_stats_accum.append(obs_subdf)
        site_stat_accum.append(site_subdf)

    site_stats = pd.concat(site_stat_accum)
    obs_stats = pd.concat(obs_stats_accum)

    from watch.utils import util_kwplot

    # dates = site_stats['start_date']
    site_stats['start_date'] = util_kwplot.fix_matplotlib_dates(site_stats['start_date'])
    site_stats['end_date'] = util_kwplot.fix_matplotlib_dates(site_stats['end_date'])
    obs_stats['observation_date'] = util_kwplot.fix_matplotlib_dates(obs_stats['observation_date'])

    phase_duration_keys = [k for k in site_stats.columns if k.startswith('duration.')]
    for phase_duration_key in phase_duration_keys:
        deltas = site_stats[phase_duration_key]
        site_stats[phase_duration_key] = util_kwplot.fix_matplotlib_timedeltas(deltas)

    import kwplot
    from watch.mlops.aggregate import hash_regions
    sns = kwplot.autosns()

    finalize_figure = util_kwplot.FigureFinalizer(
        size_inches=np.array([6.4, 4.8]) * 1.0,
        dpath=viz_dpath,
    )
    label_modifier = util_kwplot.LabelModifier()
    for phase_duration_key in phase_duration_keys:
        x = phase_duration_key
        label_modifier.add_mapping({x: x + ' (days)'})
        label_modifier.add_mapping({'rt_area': 'sqrt(area)'})

    fnum = 1

    def new_figure():
        fig = kwplot.figure(doclf=True, fnum=fnum)
        # if 0:
        #     fnum += 1
        return fig

    kwplot.close_figures()

    ### PER-SITE PLOTS
    regions = site_stats['region_id'].unique().tolist()
    region_title = hash_regions(regions)

    ax = new_figure().gca()
    sns.boxplot(data=site_stats, x='status', y='end_date', ax=ax)
    ax.set_title(f'End Date Distribution: regions={region_title}')
    ax.yaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
    ax.yaxis.set_major_locator(mdates.DayLocator(interval=356))
    finalize_figure(ax.figure, 'date_distri_stop.png')

    ax = new_figure().gca()
    sns.boxplot(data=site_stats, x='status', y='start_date', ax=ax)
    ax.set_title(f'Start Date Distribution: regions={region_title}')
    ax.yaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
    ax.yaxis.set_major_locator(mdates.DayLocator(interval=356))
    finalize_figure(ax.figure, 'date_distri_start.png')

    ax = new_figure().gca()
    ax.set_title(f'regions={region_title}')
    sns.boxplot(data=site_stats, x='status', y='rt_area', ax=ax)
    finalize_figure(ax.figure, 'status_vs_area.png')

    ax = new_figure().gca()
    sns.scatterplot(data=site_stats, x='obox_major', y='obox_minor', ax=ax, hue='status')
    ax.set_title(f'regions={region_title}')
    finalize_figure(ax.figure, 'obox_aspect_ratio.png')

    ax = new_figure().gca()
    sns.scatterplot(data=site_stats, x='obox_major', y='obox_minor', ax=ax, hue='status')
    ax.set_title(f'regions={region_title}')
    finalize_figure(ax.figure, 'obox_aspect_ratio.png')

    def corr_label(group, x, y):
        metrics_of_interest = group[[x, y]]
        metric_corr_mat = metrics_of_interest.corr()
        metric_corr = metric_corr_mat.stack()
        metric_corr.name = 'corr'
        stack_idx = metric_corr.index
        valid_idxs = [(a, b) for (a, b) in ub.unique(map(tuple, map(sorted, stack_idx.to_list()))) if a != b]
        if valid_idxs:
            metric_corr = metric_corr.loc[valid_idxs]
            # corr_lbl = 'corr({},{})={:0.4f}'.format(*metric_corr.index[0], metric_corr.iloc[0])
            corr_lbl = 'corr={:0.4f}'.format(metric_corr.iloc[0])
        else:
            corr_lbl = ''
        return corr_lbl

    phase_duration_keys = [k for k in site_stats.columns if k.startswith('duration.')]
    max_durration = site_stats[phase_duration_keys].max().max()

    import rich
    from watch.utils import util_pandas
    table = site_stats[phase_duration_keys].describe().applymap(lambda x: ''.join(str(x).partition('days')[0:2]))
    table = util_pandas.pandas_shorten_columns(table)
    rich.print(table)

    y = 'rt_area'
    ddd = site_stats[phase_duration_keys + ['site_id', 'rt_area']].melt(id_vars=['site_id', 'rt_area'])
    ddd['value'] = [getattr(d, 'days', None) for d in ddd['value']]
    ax = new_figure().gca()
    sns.scatterplot(data=ddd, x='value', y=y, ax=ax, hue='variable')
    ax.set_title(f'Durations\nregions={region_title}')
    ax.set_xscale('log')
    ax.set_yscale('log')
    for k, group in ddd.groupby('variable'):
        x = 'value'
        label_modifier.update({k: k + ': ' + corr_label(group, x, y)})
    label_modifier(ax)
    finalize_figure(ax.figure, 'area_vs_duration_melt.png')

    for k, group in ddd.groupby('variable'):
        label_modifier.update({k: k})
    ax = new_figure().gca()
    sns.boxplot(data=ddd, x='variable', y=x, ax=ax)
    ax.set_ylim(1, max_durration.days)
    label_modifier(ax)
    ax.set_title(f'Durations\nregions={region_title}')
    finalize_figure(ax.figure, 'duration_boxplot_melt.png')

    for phase_duration_key in phase_duration_keys:
        ax = new_figure().gca()
        valid_duration_df = site_stats[~pd.isnull(site_stats[phase_duration_key])].copy()
        valid_duration_df[phase_duration_key] = [d.days for d in valid_duration_df[phase_duration_key]]
        hue = 'status'
        x = phase_duration_key
        y = 'rt_area'
        for k, group in valid_duration_df.groupby(hue):
            label_modifier.update({k: k + ': ' + corr_label(group, x, y)})
            ...
        sns.scatterplot(data=valid_duration_df, x=x, y=y, ax=ax, hue=hue)
        ax.set_title(f'{phase_duration_key}\nregions={region_title}')
        ax.set_xscale('log')
        ax.set_yscale('log')
        label_modifier(ax)
        finalize_figure(ax.figure, f'area_vs_{phase_duration_key}.png')

        for k, group in valid_duration_df.groupby(hue):
            label_modifier.update({k: k})
        ax = new_figure().gca()
        sns.boxplot(data=valid_duration_df, x=hue, y=x, ax=ax)
        ax.set_ylim(1, max_durration.days)
        label_modifier(ax)
        ax.set_title(f'{phase_duration_key}\nregions={region_title}')
        finalize_figure(ax.figure, f'duration_boxplot_{phase_duration_key}.png')

    # with sns.color_palette("flare"):
    # sns.scatterplot(data=site_stats, x='rt_area', y='major_obox_ratio', hue='status')

    ### PER-OBSERVATION PLOTS

    regions = obs_stats['region_id'].unique().tolist()
    region_title = hash_regions(regions)

    ax = new_figure().gca()
    ax.set_title(f'regions={regions}')
    sns.boxplot(data=obs_stats, x='current_phase', y='observation_date', ax=ax)
    ax.yaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
    ax.yaxis.set_major_locator(mdates.DayLocator(interval=356))
    finalize_figure(ax.figure, 'phase_vs_obsdate.png')

    ax = kwplot.figure(doclf=True, fnum=4).gca()
    ax.set_title(f'regions={regions}')
    sns.boxplot(data=obs_stats, x='current_phase', y='rt_area', ax=ax)
    finalize_figure(ax.figure, 'phase_vs_size.png')

    ax = kwplot.figure(doclf=True, fnum=5).gca()
    sns.scatterplot(data=obs_stats, x='obox_major', y='obox_minor', ax=ax, hue='current_phase')
    ax.set_title(f'regions={regions}')
    finalize_figure(ax.figure, 'obox_aspect_ratio_phase.png')

    ax = kwplot.figure(doclf=True, fnum=5).gca()
    sns.scatterplot(data=obs_stats, x='obox_major', y='obox_minor', ax=ax, hue='current_phase')
    ax.set_title(f'regions={regions}')
    finalize_figure(ax.figure, 'size_verus_duration.png')

    # import xdev
    # xdev.view_diri


def geopandas_shape_stats(df):
    """
    Compute shape statistics about a geopandas dataframe (assume UTM CRS)
    """
    import kwimage
    import numpy as np
    # df['area'] = df.geometry.area
    # df['hull_area'] = df.geometry.convex_hull.area
    df['hull_rt_area'] = np.sqrt(df.geometry.convex_hull.area)
    df['rt_area'] = np.sqrt(df.geometry.area)

    obox_whs = [kwimage.MultiPolygon.from_shapely(s).oriented_bounding_box().extent
                for s in df.geometry]

    df['obox_major'] = [max(e) for e in obox_whs]
    df['obox_minor'] = [min(e) for e in obox_whs]
    df['major_obox_ratio'] = df['obox_major'] / df['obox_minor']

    # df['ch_aspect_ratio'] =
    # df['isoperimetric_quotient'] = df.geometry.apply(shapestats.ipq)
    # df['boundary_amplitude'] = df.geometry.apply(shapestats.compactness.boundary_amplitude)
    # df['eig_seitzinger'] = df.geometry.apply(shapestats.compactness.eig_seitzinger)
    return df


__config__ = GeojsonSiteStatConfig

if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/watch/cli/geojson_site_stats.py
    """
    main()



import os
import ubelt as ub
import scriptconfig as scfg


class CocoAverageFeaturesConfig(scfg.DataConfig):
    """
    Create a new kwcoco file with averaged features from multiple kwcoco files.

    High Level Steps:
        1. Load kwcoco files. Must have at least two kwcoco files.
        2. Create new kwcoco file by copying first kwcoco file.
        3. For each image ID in the kwcoco file, load the features from each kwcoco file.
            a. Average the features from each kwcoco file.
            b. Save the averaged features to the new kwcoco image.
        4. Save the new kwcoco file.

    """
    __command__ = 'average_features'
    __alias__ = ['ensemble']
    # __group__ = 'kwcoco'

    kwcoco_file_paths = scfg.Value(None,
                                   type=str,
                                   required=True,
                                   help=ub.paragraph('''
            Path to at least two kwcoco paths with predictions.
            '''),
                                   nargs='+')
    output_kwcoco_path = scfg.Value(None,
                                    type=str,
                                    required=True,
                                    help=ub.paragraph('''
            Path to the combined features kwcoco file.
            '''))
    channel_name = scfg.Value(None,
                              type=str,
                              required=True,
                              help=ub.paragraph('''
            Name of the channel in kwcoco files to merge.
            '''),
                              nargs='+', alias='channels')
    weights = scfg.Value(None,
                         type=float,
                         help=ub.paragraph('''
            Combination weight value for each prediction from kwcoco
            file. Default: All predictions are equally weighted.
            '''),
                         nargs='+')
    sensors = scfg.Value(None,
                         type=str,
                         choices=[None, 'all', 'S2', 'L8', 'WV'],
                         help=ub.paragraph('''
            Only merge channels from this type of sensor.
            '''))
    output_channel_names = scfg.Value(None,
                                      type=str,
                                      help=ub.paragraph('''
            What the name of the output channels will be after
            averaging. Needs to have the same number of channel names as
            both of the input channel names. NOTE: Channel names can be
            separated by ',' or '|' characters.
            '''))
    flexible_merge = scfg.Value(False,
                                isflag=True,
                                help=ub.paragraph('''
            If active, skip images that dont contain band when merging.
            '''))
    resolution = scfg.Value(None,
                            type=float,
                            help=ub.paragraph('''
            Set the resolution that the features will be loaded at
            and then merged.
            '''))

    io_workers = scfg.Value(
        'avail', help='number of workers used to read multiple datasets')


def split_channel_names_by_grammar(channel_names):
    """Split a string containing channel names by commas (,) and pipes (|).

    Args:
        channel_names (str):
            A string that may contain commas and pipe characters.

    Returns:
        list(str): A list of strings that were originally divided by certain characters.
    """
    import re
    channel_names = re.split(r",|\|", channel_names)
    return channel_names


def check_kwcoco_file(kwcoco_file, channel_name, sensor_names=None, flexible_merge=False):
    """Make sure that kwcoco files exist and contain required channel name.

    Args:
        kwcoco_file (kwcoco.CocoDataset): kwcoco file containing images.
        channel_name (str): Name of channel thats required to be in kwcoco file.
        sensor_names (list(str), optional): Only check images of from these types of sensors. Defaults to None.
        flexible_merge (bool, optional): Skip images that do not contain channel_name. Defaults to False.

    Returns:
        missing_image_names (list): A list of names corresponding to images without the channel name.
    """
    import kwcoco

    # Get all images in kwcoco file.
    images: kwcoco.coco_dataset.Videos = kwcoco_file.images()

    n_all_images = len(images)
    if sensor_names is not None:
        # Filter to only images with a chosen sensor
        flags = [s in sensor_names for s in images.lookup("sensor_coarse", None)]
        images = images.compress(flags)
        print(f"INFO: Number of {sensor_names} images: [{len(images)}/{n_all_images}]")

        if len(images) == 0:
            raise ValueError(f"No images in {kwcoco_file.fpath} containing images of sensor {sensor_names}.")

    # Breakup channels names based on , and | characters and convert channel_name from str to list.
    channel_names = split_channel_names_by_grammar(channel_name)

    available_image_names, missing_image_names = [], []
    for coco_img in images.coco_images:
        # Check if all requested channel names were found in image.
        channel_matches = []
        merged_image_channel_names = ",".join(list(coco_img.channels.keys()))
        image_channels = split_channel_names_by_grammar(merged_image_channel_names)
        for channel_name in channel_names:
            channel_matches.append(channel_name in image_channels)

        # If all channel matches are true then skip.
        image_name = kwcoco_file.index.imgs[coco_img.img["id"]]["name"]
        if not all(channel_matches):
            missing_image_names.append(image_name)
            if flexible_merge is False:
                raise AssertionError(
                    f"Channel(s) '{channel_names}' not found in image {coco_img.img['id']} of kwcoco file {kwcoco_file.fpath}. Only channels found: {coco_img.channels}"
                )
        else:
            available_image_names.append(image_name)

    if flexible_merge:
        print(kwcoco_file.fpath)
        print(
            f"INFO: Number of images without requested channel names from [{len(missing_image_names)}/{len(images.coco_images)}]"
        )

        # Make sure that at least one image contains all requested channel names.
        if len(missing_image_names) == len(images.coco_images):
            raise ValueError(f"All of the images in {kwcoco_file.fpath} were missing channel(s) {channel_names}")

    return available_image_names, missing_image_names


def merge_kwcoco_channels(kwcoco_file_paths,
                          output_kwcoco_path,
                          channel_names,
                          weights,
                          output_channel_names,
                          sensor_names=None,
                          resolution=None,
                          flexible_merge=False,
                          io_workers='avail'):
    """
    Compute a weighted mean of channels from separate kwcoco file and save into
    merged kwcoco file.

    Assumptions:
        - The channel_nams to merge are not a subset of a group of channels targeted in the kwcoco file.
            - I.e. 'salient' in 'salient|notsalient' will not work.

    Args:
        kwcoco_file_paths (list(str)):
            A list of paths representing pathes to kwcoco files to be merged.

        output_kwcoco_path (str):
            Local path to the kwcoco file with merged channels.

        channel_names (list(str)):
            A list of channel names corresponding to the channel name to merge
            from each kwcoco file. Note, the length of the channel names be
            equal to the number of kwcoco file paths.

        weights (list(int)):
            A list of floats representing how much weight a particular kwcoco
            file should contribute to the final merged prediction.

        output_channel_names (list(str)):
            A list containing the names of the output channel names. Must
            contain the same number of channel names as the input channel
            names.

        sensor_names (list(str), optional):
            Only merge images belonging to sensors in this list. Defaults to None (aka do not filter by sensor).

        resolution (int | str, optional):
            GSD to resize the resolution of the images to. Defaults to None.

        flexible_merge (bool, optional):
            Skip images that do not contain channel_name. Defaults to False.

    Example:
        >>> # TEST 1: Merge two kwcoco files with the same number of images and plot results.
        >>> from watch.cli.coco_average_features import *  # NOQA
        >>> import watch
        >>> import kwimage
        >>> import numpy as np
        >>> from kwcoco.demo.perterb import perterb_coco
        >>> import kwcoco
        >>> dpath = ub.Path.appdir('watch/test/coco_average_features')
        >>> base_dset = watch.coerce_kwcoco('watch-msi', geodata=True, dates=True, image_size=(64, 64), num_videos=2, num_frames=2)
        >>> # Construct two copies of the same data with slightly different heatmaps
        >>> dset1 = perterb_coco(base_dset.copy(), box_noise=0.5, cls_noise=0.5, n_fp=10, n_fn=10, rng=32)
        >>> dset2 = base_dset.copy()
        >>> for video in dset1.dataset['videos']:
        ...      video['resolution'] = '10GSD'
        >>> for video in dset2.dataset['videos']:
        ...      video['resolution'] = '10GSD'
        >>> dset1.fpath = ub.Path(dset1.fpath).augment(stemsuffix='_heatmap1')
        >>> dset2.fpath = ub.Path(dset2.fpath).augment(stemsuffix='_heatmap2')
        >>> watch.demo.smart_kwcoco_demodata.hack_in_heatmaps(dset1, heatmap_dname='dummy_heatmap1', with_nan=0, rng=423432)
        >>> watch.demo.smart_kwcoco_demodata.hack_in_heatmaps(dset2, heatmap_dname='dummy_heatmap2', with_nan=0, rng=132129)
        >>> dset1.dump(dset1.fpath)
        >>> dset2.dump(dset2.fpath)
        >>> # Build method args
        >>> kwcoco_file_paths = [dset1.fpath, dset2.fpath]
        >>> output_bundle_dpath = (dpath / 'merge_bundle').delete().ensuredir()
        >>> output_kwcoco_path = output_bundle_dpath / 'data.kwcoco.json'
        >>> channel_names = ['notsalient|salient'] * 2
        >>> weights = [1.0, 1.0]
        >>> output_channel_names = 'notsalient|salient'
        >>> sensor_name = None
        >>> resolution = '12GSD'
        >>> # Execute merge
        >>> merge_kwcoco_channels(kwcoco_file_paths, output_kwcoco_path,
        >>>                       channel_names, weights, output_channel_names,
        >>>                       sensor_name, resolution=resolution)
        >>> # Check results
        >>> output_dset = kwcoco.CocoDataset(output_kwcoco_path)
        >>> gid = output_dset.images()[1]
        >>> imdata1 = dset1.coco_image(gid).imdelay('salient', space='asset').finalize()
        >>> imdata2 = dset2.coco_image(gid).imdelay('salient', space='asset').finalize()
        >>> imdataM = output_dset.coco_image(gid).imdelay('salient', space='asset').finalize()
        >>> imdata1_img = dset1.coco_image(gid).imdelay('salient', space='image').finalize()
        >>> imdata2_img = dset2.coco_image(gid).imdelay('salient', space='image').finalize()
        >>> imdataM_img = output_dset.coco_image(gid).imdelay('salient', space='image').finalize()
        >>> print(f'Weights: {weights}')
        >>> print(f'Img1  mean: {np.nan_to_num(imdata1).mean()}')
        >>> print(f'Img2  mean: {np.nan_to_num(imdata2).mean()}')
        >>> print(f'Merge mean: {np.nan_to_num(imdataM).mean()}')
        >>> print()
        >>> print(f'Img1  shape (asset space): {imdata1.shape}')
        >>> print(f'Img2  shape (asset space): {imdata2.shape}')
        >>> print(f'Merge shape (asset space): {imdataM.shape}')
        >>> print(f'Img1  shape (img space): {imdata1_img.shape}')
        >>> print(f'Img2  shape (img space): {imdata2_img.shape}')
        >>> print(f'Merge shape (img space): {imdataM_img.shape}')
        >>> os.remove(dset1.fpath)
        >>> os.remove(dset2.fpath)
        >>> os.remove(output_dset.fpath)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> F = kwimage.fill_nans_with_checkers
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdata1)), title='img1 (asset)', pnum=(2, 3, 1), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdata2)), title='img2 (asset)', pnum=(2, 3, 2), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdataM)), title='mean (asset)', pnum=(2, 3, 3), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdata1_img)), title='img1 (img)', pnum=(2, 3, 4), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdata2_img)), title='img2 (img)', pnum=(2, 3, 5), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdataM_img)), title='mean (img)', pnum=(2, 3, 6), fnum=1)
        >>> save_figure_path = dpath / 'test_1_result_plot.png'
        >>> import matplotlib.pyplot as plt
        >>> plt.savefig(save_figure_path)
        >>> print(f'Test 1 plot saved to: {save_figure_path}')

    Example:
        >>> # TEST 2: Merge two kwcoco files with geo information.
        >>> from watch.cli.coco_average_features import *  # NOQA
        >>> import watch
        >>> import kwimage
        >>> from kwcoco.demo.perterb import perterb_coco
        >>> import kwcoco
        >>> import numpy as np
        >>> dpath = ub.Path.appdir('watch/test/coco_average_features')
        >>> base_dset = watch.coerce_kwcoco('watch-msi', geodata=True, dates=True, image_size=(64, 64), num_videos=2, num_frames=2)
        >>> # Construct two copies of the same data with slightly different heatmaps
        >>> dset1 = perterb_coco(base_dset.copy(), box_noise=0.5, cls_noise=0.5, n_fp=10, n_fn=10, rng=32)
        >>> dset2 = base_dset.copy()
        >>> dset1.fpath = ub.Path(dset1.fpath).augment(stemsuffix='_heatmap1')
        >>> dset2.fpath = ub.Path(dset2.fpath).augment(stemsuffix='_heatmap2')
        >>> watch.demo.smart_kwcoco_demodata.hack_in_heatmaps(dset1, heatmap_dname='dummy_heatmap1', with_nan=0, rng=423555)
        >>> watch.demo.smart_kwcoco_demodata.hack_in_heatmaps(dset2, heatmap_dname='dummy_heatmap2', with_nan=0, rng=132666)
        >>> dset1.dump(dset1.fpath)
        >>> dset2.dump(dset2.fpath)
        >>> # Build method args
        >>> kwcoco_file_paths = [dset1.fpath, dset2.fpath]
        >>> output_bundle_dpath = (dpath / 'merge_bundle').delete().ensuredir()
        >>> output_kwcoco_path = output_bundle_dpath / 'data.kwcoco.json'
        >>> channel_names = ['notsalient|salient'] * 2
        >>> weights = [1.0, 1.0]
        >>> output_channel_names = 'notsalient|salient'
        >>> sensor_name = None
        >>> resolution = None
        >>> # Execute merge
        >>> merge_kwcoco_channels(kwcoco_file_paths, output_kwcoco_path,
        >>>                       channel_names, weights, output_channel_names,
        >>>                       sensor_name, resolution=resolution)
        >>> # Check results
        >>> output_dset = kwcoco.CocoDataset(output_kwcoco_path)
        >>> gid = 3
        >>> imdata1 = dset1.coco_image(gid).imdelay('salient', space='asset').finalize()
        >>> imdata2 = dset2.coco_image(gid).imdelay('salient', space='asset').finalize()
        >>> imdataM = output_dset.coco_image(gid).imdelay('salient', space='asset').finalize()
        >>> imdata1_img = dset1.coco_image(gid).imdelay('salient', space='image').finalize()
        >>> imdata2_img = dset2.coco_image(gid).imdelay('salient', space='image').finalize()
        >>> imdataM_img = output_dset.coco_image(gid).imdelay('salient', space='image').finalize()
        >>> print(f'Weights: {weights}')
        >>> print(f'Img1  mean: {np.nan_to_num(imdata1).mean()}')
        >>> print(f'Img2  mean: {np.nan_to_num(imdata2).mean()}')
        >>> print(f'Merge mean: {np.nan_to_num(imdataM).mean()}')
        >>> print()
        >>> print(f'Img1  shape (asset space): {imdata1.shape}')
        >>> print(f'Img2  shape (asset space): {imdata2.shape}')
        >>> print(f'Merge shape (asset space): {imdataM.shape}')
        >>> print(f'Img1  shape (img space): {imdata1_img.shape}')
        >>> print(f'Img2  shape (img space): {imdata2_img.shape}')
        >>> print(f'Merge shape (img space): {imdataM_img.shape}')
        >>> os.remove(dset1.fpath)
        >>> os.remove(dset2.fpath)
        >>> os.remove(output_dset.fpath)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> F = kwimage.fill_nans_with_checkers
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdata1)), title='img1 (asset)', pnum=(2, 3, 1), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdata2)), title='img2 (asset)', pnum=(2, 3, 2), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdataM)), title='mean (asset)', pnum=(2, 3, 3), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdata1_img)), title='img1 (img)', pnum=(2, 3, 4), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdata2_img)), title='img2 (img)', pnum=(2, 3, 5), fnum=1)
        >>> kwplot.imshow(F(kwimage.normalize_intensity(imdataM_img)), title='mean (img)', pnum=(2, 3, 6), fnum=1)
        >>> save_figure_path = dpath / 'test_2_result_plot.png'
        >>> import matplotlib.pyplot as plt
        >>> plt.savefig(save_figure_path)
        >>> print(f'Test 2 plot saved to: {save_figure_path}')


    Ignore:
        import xdev
        globals().update(xdev.get_func_kwargs(merge_kwcoco_channels))
    """
    from watch.utils.kwcoco_extensions import transfer_geo_metadata
    from watch.tasks.fusion.coco_stitcher import quantize_image
    from watch import exceptions
    from tqdm import tqdm
    import kwimage
    import kwarray
    import kwcoco
    import numpy as np

    # Check args.
    if len(kwcoco_file_paths) <= 1:
        raise ValueError(f"Need at least 2 kwcoco files to merge and only recieved {len(kwcoco_file_paths)}")

    ## If weights argument is not None then make sure it is the same length as kwcoco files.
    if weights is not None:
        if len(weights) != len(kwcoco_file_paths):
            raise ValueError(
                f"If weights is not None, number of weights ({len(weights)}) must be equal to number of kwcoco files ({len(kwcoco_file_paths)})."
            )
    else:
        # Set weight of all files to 1.
        weights = [1] * len(kwcoco_file_paths)

    ## If channels is not one value then make sure that it has the same number of values as kwcoco files.
    if len(channel_names) != 1:
        if len(channel_names) != len(kwcoco_file_paths):
            raise ValueError(
                f"If more than one channel name, number of channel names ({len(channel_names)}) must be equal to number of kwcoco files ({len(kwcoco_file_paths)})."
            )
        # Make sure the number of channel names per file are equal.
        n_channels_per_file = []
        for kwcoco_file_channel_names in channel_names:
            n_channels_per_file.append(len(split_channel_names_by_grammar(kwcoco_file_channel_names)))

        if len(set(n_channels_per_file)) != 1:
            raise ValueError(f"Number of requested channels per kwcoco file are not equal: {n_channels_per_file}")
    else:
        channel_names = channel_names * len(kwcoco_file_paths)

    ## If no merge_channel_name given then use first
    if output_channel_names is None:
        output_channel_names = channel_names[0]
        print(f"INFO: No output channel name given, using channel name: {channel_names[0]}")
    else:
        # Make sure that the size of output channel names and input channel names are equal in length.
        split_output_channel_names = split_channel_names_by_grammar(output_channel_names)

        # Get the number of channels per kwcoco file.
        n_channels_per_file = []
        for kwcoco_file_channel_names in channel_names:
            n_channels_per_file.append(len(split_channel_names_by_grammar(kwcoco_file_channel_names)))

        if len(split_output_channel_names) != list(set(n_channels_per_file))[0]:
            raise ValueError(
                f"Number of output channels ({(output_channel_names)}) does not match number of requested channels ({list(set(n_channels_per_file))[0]})."
            )

    # Load and merge images from kwcoco files.
    ## Load kwcoco files.
    # kwcoco_files = [kwcoco.CocoDataset.coerce(p) for p in kwcoco_file_paths]
    kwcoco_files = list(kwcoco.CocoDataset.coerce_multiple(
        kwcoco_file_paths, workers=io_workers))

    ## Check kwcoco files to see that they exist and contain the required channels.
    all_available_image_names, all_missing_image_names = [], []
    for kwcoco_file_path, channel_name in zip(kwcoco_files, channel_names):
        available_image_names, missing_image_names = check_kwcoco_file(kwcoco_file_path,
                                                                       channel_name,
                                                                       sensor_names=sensor_names,
                                                                       flexible_merge=flexible_merge)
        all_missing_image_names.append(set(missing_image_names))
        all_available_image_names.append(set(available_image_names))

    ## Find common image names.
    all_image_names = list(set.union(*all_available_image_names))
    common_image_names = list(set.intersection(*all_available_image_names))
    missing_image_names = list(set.difference(set(common_image_names), set(all_image_names)))
    if len(common_image_names) == 0:
        raise ValueError("No common images found between all kwcoco files.")

    if len(missing_image_names) != 0:
        print(f'Out of all images, {len(missing_image_names)} are missing from at least one kwcoco file.')

    # Create output kwcoco by copying first kwcoco file.
    merge_kwcoco = kwcoco_files[0].copy()

    ## Remove missing images from kwcoco file.
    merge_kwcoco.remove_images(missing_image_names)

    # Load channel images from each viable image_id.
    output_kwcoco_path = ub.Path(output_kwcoco_path)
    save_assest_dir = (output_kwcoco_path.parent / "_assets").ensuredir()

    # Get channel names for each file.
    kwcoco_channel_names = []
    for channel_name in channel_names:
        split_channel_names = split_channel_names_by_grammar(channel_name)
        combined_channel_names = kwcoco.FusedChannelSpec.coerce(split_channel_names)
        kwcoco_channel_names.append(combined_channel_names)

    # Generate output channel names in (channel_name1|channel_name2|channel_name3) format.
    output_channels = kwcoco.FusedChannelSpec.coerce(output_channel_names)

    pbar = tqdm(merge_kwcoco.index.imgs.items(), desc="Merging images", colour="green")
    for image_id, image_info in pbar:
        # Skip this image if its name is in the missing image name list.
        image_name = merge_kwcoco.index.imgs[image_id]["name"]
        if image_name not in common_image_names:
            continue

        # If sensor name specified, only merge channels for images from this sensor.
        if sensor_names is not None:
            if image_info["sensor_coarse"] not in sensor_names:
                continue

        # Get the merged kwcoco image.
        merge_coco_img = merge_kwcoco.coco_image(image_id)

        # If the target resolution is None, then set the target resolution as the image with the highest resolution.
        if resolution is None:
            asset_resolutions = []
            for kwcoco_file, channel_name in zip(kwcoco_files, kwcoco_channel_names):
                asset_mag = kwcoco_file.coco_image(image_id).resolution(channel=channel_name, space='asset')['mag']
                asset_resolutions.append(asset_mag[0])

            # ASSUMPTION: Scales are symmetric.
            index = np.argmin([mag for mag in asset_resolutions])
            img_resolution = asset_resolutions[index]

        # Get asset channels from each kwcoco file based on image_name.
        gathered_parts = []
        for kwcoco_file, channel_name in zip(kwcoco_files, kwcoco_channel_names):
            # Get the kwcoco specific image id from image name.
            kwfile_image_id = kwcoco_file.index.name_to_img[image_name]["id"]

            # Get the corresponding asset within the kwcoco image file.
            coco_img = kwcoco_file.coco_image(kwfile_image_id)
            asset = coco_img.find_asset_obj(channel_name)

            # Load the image in 'video' space to ensure consistent space for target resolution.
            delayed = coco_img.imdelay(channel_name, space='video', resolution=resolution)
            gathered_parts.append({
                'asset': asset,
                'delayed': delayed,
            })

        # Check that assets are the same dimentions and have the same transforms.
        if __debug__:
            # Assuming a high degree of alignment between input assets will
            # need to extend if this assumption breaks.
            assert ub.allsame([p['asset']['width'] for p in gathered_parts])
            assert ub.allsame([p['asset']['height'] for p in gathered_parts])
            warps = [p['asset']['warp_aux_to_img'] for p in gathered_parts]
            assert ub.allsame(warps)

        # Apply a weight mask to each image while masking out NaN values.
        accum = kwarray.Stitcher(gathered_parts[0]['delayed'].shape)
        for part, weight in zip(gathered_parts, weights):
            delayed = part['delayed']
            image_data = delayed.finalize(nodata_method='float')
            pxl_weight = (1 - np.isnan(image_data)) * weight
            accum.add((slice(None), slice(None)), image_data, pxl_weight)

        average_image_data = accum.finalize()

        # DEBUG: Plot the original and merged images.
        if 0:
            import kwplot
            kwplot.autompl()
            a = gathered_parts[0]['delayed']
            b = gathered_parts[1]['delayed']
            im1 = a.finalize(nodata_method='float')[..., 1]
            im2 = b.finalize(nodata_method='float')[..., 1]
            canvas = kwimage.stack_images([im1, im2, average_image_data[..., 1]], axis=1)
            kwplot.imshow(canvas)

        # TODO: better backend name.
        path_chan = output_channels.path_sanitize()
        img_name = merge_coco_img.img.get("name", "")
        average_fname = f"merged_{img_name}_{path_chan}.tif"
        average_fpath = save_assest_dir / average_fname

        # Check if there is already an asset with the same channel_names.
        output_obj = merge_coco_img.find_asset_obj(output_channels)

        # TODO: use the CocoSticher here

        # Find the transformation from target to image space.
        if resolution is None:
            scale_target_from_vid = kwimage.Affine.scale(
                coco_img._scalefactor_for_resolution(space='asset',
                                                     channel=output_channel_names,
                                                     resolution=img_resolution))
        else:
            scale_target_from_vid = kwimage.Affine.scale(
                coco_img._scalefactor_for_resolution(space='video', channel=output_channel_names,
                                                     resolution=resolution))

        warp_target_from_img = scale_target_from_vid @ coco_img.warp_vid_from_img
        warp_img_from_target = warp_target_from_img.inv()

        # TODO: probably want to use the CocoStitcher directly.
        # TODO: if we know the min/max range, specify it.
        quant_data, quantization = quantize_image(average_image_data)

        if output_obj is not None:
            # Overwrite the data in the output auxiliary item.
            output_obj["file_name"] = os.fspath(average_fpath)
            output_obj['height'] = average_image_data.shape[0]
            output_obj['width'] = average_image_data.shape[1]
            if len(average_image_data.shape) > 2:
                # average_image_data: (H, W, C)
                output_obj['num_bands'] = average_image_data.shape[-1]
            else:
                # average_image_data: (H, W)
                output_obj['num_bands'] = 1
            output_obj['warp_aux_to_img'] = warp_img_from_target.concise()
            output_obj['quantization'] = quantization

        # Get default kwargs from ../tasks/fusion/predict.py:CocoStitchingManager.finalize_image
        write_kwargs = {}
        write_kwargs['blocksize'] = 128
        write_kwargs['compress'] = 'DEFLATE'

        # TODO: Possibly add `wld_crs_info` to write_kwargs.

        # Write the averaged image data
        kwimage.imwrite(average_fpath, quant_data, backend="gdal", nodata=quantization['nodata'], **write_kwargs)

        # Update all channels with projection info.
        try:
            transfer_geo_metadata(merge_kwcoco, image_id)
        except exceptions.GeoMetadataNotFound as ex:
            print("warning ex = {!r}".format(ex))
            pass

    # Save kwcoco file.
    # merge_kwcoco.validate()
    merge_kwcoco.dump(output_kwcoco_path)
    print(f"Saved merged kwcoco file to: {output_kwcoco_path}")


def main(cmdline=True, **kw):
    """
    Main function for merge_kwcoco_channels.
    See :class:``CocoAverageFeaturesConfig` for details

    TODO: Add examples

    Example:
        >>> from watch.cli.coco_average_features import *  # NOQA
        >>> import watch
        >>> import kwimage
        >>> import numpy as np
        >>> from kwcoco.demo.perterb import perterb_coco
        >>> import kwcoco
        >>> dpath = ub.Path.appdir('watch/test/coco_average_features_main')
        >>> base_dset = watch.coerce_kwcoco('watch-msi', geodata=True, dates=True, image_size=(64, 64), num_videos=2, num_frames=2)
        >>> # Construct two copies of the same data with slightly different heatmaps
        >>> dset1 = perterb_coco(base_dset.copy(), box_noise=0.5, cls_noise=0.5, n_fp=10, n_fn=10, rng=32)
        >>> dset2 = base_dset.copy()
        >>> for video in dset1.dataset['videos']:
        ...      video['resolution'] = '10GSD'
        >>> for video in dset2.dataset['videos']:
        ...      video['resolution'] = '10GSD'
        >>> dset1.fpath = ub.Path(dset1.fpath).augment(stemsuffix='_heatmap1')
        >>> dset2.fpath = ub.Path(dset2.fpath).augment(stemsuffix='_heatmap2')
        >>> watch.demo.smart_kwcoco_demodata.hack_in_heatmaps(dset1, heatmap_dname='dummy_heatmap1', with_nan=0, rng=423432)
        >>> watch.demo.smart_kwcoco_demodata.hack_in_heatmaps(dset2, heatmap_dname='dummy_heatmap2', with_nan=0, rng=132129)
        >>> dset1.dump(dset1.fpath)
        >>> dset2.dump(dset2.fpath)
        >>> output_kwcoco_path = dpath / 'output.kwcoco.zip'
        >>> # Execute merge
        >>> kwargs = {
        >>>     'kwcoco_file_paths': [dset1.fpath, dset2.fpath],
        >>>     'output_kwcoco_path': output_kwcoco_path,
        >>>     'channels': ['notsalient|salient'],
        >>>     'resolution': 30,
        >>> }
        >>> cmdline = 0
        >>> main(cmdline=cmdline, **kwargs)
        >>> output_dset = kwcoco.CocoDataset(output_kwcoco_path)
        >>> coco_img = output_dset.images().coco_images[0]
        >>> import numpy as np
        >>> assert not np.all(np.isnan(coco_img.imdelay('salient').finalize()))
    """
    config = CocoAverageFeaturesConfig.cli(
        data=kw, cmdline=cmdline, strict=True)
    import rich
    rich.print(ub.urepr(config))

    # Merge kwcoco files along certain channels.
    merge_kwcoco_channels(
        config['kwcoco_file_paths'],
        config['output_kwcoco_path'],
        config['channel_name'],
        config['weights'],
        config['output_channel_names'],
        sensor_names=config['sensors'],
        resolution=config['resolution'],
        flexible_merge=config['flexible_merge'],
        io_workers=config.io_workers,
    )


__config__ = CocoAverageFeaturesConfig


if __name__ == "__main__":
    main()



import os
import json

from watch.tasks.rutgers_material_change_detection.utils.util_misc import get_repo_dir


def get_dataset_root_dir(dataset_name):
    # Load dataset directory json file.
    base_repo_dir = get_repo_dir()
    dset_dir_json_file_path = os.path.join(base_repo_dir, "transformer", "datasets", "dataset_directories.json")

    ## Check that json file exists.
    if os.path.isfile(dset_dir_json_file_path) is False:
        raise FileNotFoundError(f'Dataset direct json file not found at: "{dset_dir_json_file_path}"')

    dataset_dirs = json.load(open(dset_dir_json_file_path, "r"))

    # Get dataset directory.
    try:
        dset_root_dir = dataset_dirs[dataset_name]
    except KeyError:
        raise KeyError(
            f'Dataset name "{dataset_name}" not found in dataset directory json file at "{dset_dir_json_file_path}"'
        )

    return dset_root_dir


def get_base_paths(key_name):
    # Load base paths json file.
    base_repo_dir = get_repo_dir()
    json_file_path = os.path.join(base_repo_dir, "base_paths.json")

    ## Check that json file exists.
    if os.path.isfile(json_file_path) is False:
        raise FileNotFoundError(f'Base paths json file not found at: "{json_file_path}"')

    base_paths = json.load(open(json_file_path, "r"))

    # Get dataset directory.
    try:
        base_path = base_paths[key_name]
    except KeyError:
        raise KeyError(f'Base path name "{key_name}" not found in json file at "{base_paths}"')

    return base_path



"""
In the interest of development speed we can't always be perfect about correctly
passing the right metadata to functions that need it. This module serves as a
place to store hard-coded heuristics so we are explicit about where we are
cutting corners or using magic numbers. The idea is that  this will make it
easier for us to go back and make the code robust.

References:
    https://gitlab.kitware.com/smart/annotations-wiki/-/blob/main/Annotation-Status-Types.md

"""
import ubelt as ub

HEURISTIC_START_STATES = {
    'No Activity',
}

HEURISTIC_END_STATES = {
    'Post Construction'
}


# # FIXME: Hard-coded category aliases.
# https://smartgitlab.com/TE/standards
# # The correct way to handle these would be to have some information in the
# # kwcoco category dictionary that specifies how the categories should be
# # interpreted.
# _HEURISTIC_CATEGORIES = {
#     'background': {'background', 'No Activity', 'Post Construction'},
#     'pre_background': {'No Activity'},
#     'post_background': {'Post Construction'},
#     'ignore': {'ignore', 'Unknown', 'clouds'},
# }


# TODO: ensure consistency with IARPA?
# https://smartgitlab.com/TE/annotations/-/wikis/Annotation-Status-Types
# https://smartgitlab.com/TE/metrics-and-test-framework/-/blob/main/iarpa_smart_metrics/evaluation.py#L1205
# NOTE: A "Status" is not a category.
# It indicates what sort of annotation detail is available.
HUERISTIC_STATUS_DATA = [
    {'name': 'seen', 'color': 'cyan'},
    {'name': 'train', 'color': 'cyan'},

    {'name': 'ignore', 'color': 'lightsalmon'},

    # Note: 'colors for these status labels are undefined, using neutral gray
    {'name': 'negative', 'color': 'gray'},
    {'name': 'negative_unbounded', 'color': 'gray'},

    {'name': 'positive_excluded', 'color': 'gray'},

    {'name': 'positive_annotated', 'color': 'black'},
    {'name': 'positive_annotated_static', 'color': 'black'},
    {'name': 'positive_partial', 'color': 'black'},
    {'name': 'positive_pending', 'color': 'black'},

    {'name': 'positive_unbounded', 'color': 'darkviolet'},

    # TODO? Add alias of pending for "positive_pending"? For QFabric?

    {'name': 'system_confirmed', 'color': 'kitware_blue'},
]


# "Positive Match Confusion" is the label the truth is given when it has some
# match in our set of positive predictions.  Denote what type of confusion a
# truth status incurs when it is matched.
PHASE_STATUS_TO_MATCHED_CONFUSION = {
    'seen'                      : 'gt_seen',
    'train'                     : 'gt_seen',

    'ignore'                    : 'gt_ignore',

    'negative'                  : 'gt_false_pos',
    'negative_unbounded'        : 'gt_false_pos',

    'positive_excluded'         : 'gt_false_pos',

    'positive_annotated'        : 'gt_true_pos',
    'positive_annotated_static' : 'gt_true_pos',
    'positive_partial'          : 'gt_true_pos',
    'positive_pending'          : 'gt_true_pos',

    'positive_unbounded'        : 'gt_positive_unbounded',

}

# Mapping of annotation status to the kwcoco category name
# Used in project annotations
PHASE_STATUS_TO_KWCOCO_CATNAME = {
    'seen'                     : None,
    'train'                    : None,

    'ignore'                    : 'ignore',

    'negative'                  : 'negative',
    'negative_unbounded'        : 'negative',

    'positive_annotated'        : None,  # This must have a category already do not map
    'positive_annotated_static' : None,  # This must have a category already do not map
    'positive_excluded'         : 'ignore',  # This is positive, but is not "big" enough
    'positive_partial'          : 'positive',  # Does not have phase labels
    'positive_pending'          : 'positive',  # Does not have phase labels

    'positive_unbounded'        : 'positive',  # Start or end date might not be defined
}

IARPA_STATUS_TO_INFO = {row['name']: row for row in HUERISTIC_STATUS_DATA}

# update HUERISTIC_STATUS_DATA
for name, row in IARPA_STATUS_TO_INFO.items():
    if name in PHASE_STATUS_TO_KWCOCO_CATNAME:
        row['kwcoco_catname'] = PHASE_STATUS_TO_KWCOCO_CATNAME[name]

for name, row in IARPA_STATUS_TO_INFO.items():
    if name in PHASE_STATUS_TO_MATCHED_CONFUSION:
        row['positive_match_confusion'] = PHASE_STATUS_TO_MATCHED_CONFUSION[name]


if 0:
    import pandas as pd
    print(pd.DataFrame(HUERISTIC_STATUS_DATA).to_string())


IARPA_REAL_STATUS = {
    'positive': ["positive_annotated", "positive_annotated_static", "positive_partial", "positive_pending"],
    'negative': ["positive_excluded", "negative", "negative_unbounded"],
    'ignore': 'ignore',
}


IARPA_CONFUSION_COLORS = {}
IARPA_CONFUSION_COLORS['gt_true_neg'] = 'darkgreen'  # no IARPA color for this, make one up.
IARPA_CONFUSION_COLORS['gt_true_pos'] = 'lime'
IARPA_CONFUSION_COLORS['gt_false_pos'] = 'red'
IARPA_CONFUSION_COLORS['gt_false_neg'] = 'black'
IARPA_CONFUSION_COLORS['gt_positive_unbounded'] = "darkviolet"
IARPA_CONFUSION_COLORS['gt_ignore'] = "lightsalmon"
IARPA_CONFUSION_COLORS['gt_seen'] = "gray"
IARPA_CONFUSION_COLORS['sm_pos_match'] = "orange"
IARPA_CONFUSION_COLORS['sm_partially_wrong'] = "aquamarine"
IARPA_CONFUSION_COLORS['sm_completely_wrong'] = "magenta"


def iarpa_assign_truth_confusion(truth_status, has_positive_match):
    """
    Example:
        >>> from watch.heuristics import *  # NOQA
        >>> import pandas as pd
        >>> rows = []
        >>> for truth_status in IARPA_STATUS_TO_INFO.keys():
        >>>     for has_positive_match in [0, 1]:
        >>>         gt_cfsn = iarpa_assign_truth_confusion(truth_status, has_positive_match)
        >>>         rows.append({
        >>>             'truth_status': truth_status,
        >>>             'has_positive_match': has_positive_match,
        >>>             'confusion': gt_cfsn,
        >>>         })
        >>> print(pd.DataFrame(rows).to_string())
    """
    gt_cfsn = None
    if has_positive_match:
        gt_cfsn = IARPA_STATUS_TO_INFO[truth_status].get('positive_match_confusion', None)
    else:
        if truth_status in ["positive_unbounded"]:
            gt_cfsn = 'gt_positive_unbounded'
        elif truth_status in ["ignore"]:
            gt_cfsn = 'gt_ignore'
        elif truth_status in ['seen', 'train']:
            gt_cfsn = 'gt_seen'
        elif truth_status in IARPA_REAL_STATUS['negative']:
            gt_cfsn = 'gt_true_neg'
        elif truth_status in IARPA_REAL_STATUS['positive']:
            gt_cfsn = 'gt_false_neg'

    return gt_cfsn


def iarpa_assign_pred_confusion(truth_match_statuses):
    """
    Example:
        >>> from watch.heuristics import *  # NOQA
        >>> import itertools as it
        >>> truth_match_statuses = {'positive_partial', 'positive_excluded'}
        >>> for combo in it.combinations(IARPA_STATUS_TO_INFO, 2):
        >>>     truth_match_statuses = combo
        >>>     pred_cfsn = iarpa_assign_pred_confusion(truth_match_statuses)
        >>>     print(f'{pred_cfsn=} for {truth_match_statuses}')
    """
    pred_cfsn = None
    if not truth_match_statuses:
        pred_cfsn = 'sm_completely_wrong'

    truth_cfsns = {
        IARPA_STATUS_TO_INFO[s].get('positive_match_confusion', None)
        for s in truth_match_statuses
    }
    if 'gt_true_pos' in truth_cfsns:
        if 'gt_false_pos' in truth_cfsns:
            pred_cfsn = 'sm_partially_wrong'
        else:
            pred_cfsn = 'sm_pos_match'
    elif 'gt_false_pos' in truth_cfsns:
        pred_cfsn = 'sm_completely_wrong'

    return pred_cfsn


# metrics-and-test-framework/evaluation.py:1684
# Note: the condition field is in development the idea is to
# encode when a category should be added as a label.

def TAG_IF(tag, condition):
    return {'type': 'condition', 'op': 'tag_if', 'tag': tag, 'condition': condition}


def CONDITION(op, args):
    return {'type': 'condition', 'op': op, 'args': args}


def ALL(*args):
    return {'type': 'condition', 'op': 'all', 'args': args}


CATEGORIES = [
    # TODO: clouds
    {
        'name': 'background',
        'color': 'black',
        'scored': False,
        'required': True,
        'tags': ['background']
    },

    # 'color' lightsalmon
    {
        'name': 'ignore',
        'color': 'violet',
        'scored': False,
        'required': True,
        'tags': ['ignore']
    },

    # {'name': 'clouds', 'color': 'offwhite', 'scored': False},
    {
        'name': 'Unknown',
        'color': 'blueviolet',
        'scored': False,
        'tags': ['ignore'],
    },

    {
        'name': 'positive',
        'color': 'palegreen',
        'scored': False,
        # 'conditional_tags': [
        #     TAG_IF('positive', CONDITION('TASK_EQ', 'saliency')),
        #     TAG_IF('ignore', CONDITION('TASK_EQ', 'class')),
        # ],
    },

    {
        'name': 'negative',
        'color': 'gray',
        'scored': False,
        'tags': ['background'],
        # 'conditional_tags': [
        #     TAG_IF('hard_negative', CONDITION('TASK_EQ', 'class')),
        # ],
    },

    {
        'name': 'Site Preparation',
        'color': 'gold',
        'scored': True,
        'tags': ['positive'],
    },
    {
        'name': 'Active Construction',
        'color': 'lime',
        'scored': True,
        'tags': ['positive'],
    },
    # Conditional classes
    {
        'name': 'Post Construction',
        'color': 'darkturquoise',
        'scored': True,
        'tags': ['positive'],
        # 'conditional_tags': [
        #     TAG_IF('background', CONDITION('TASK_EQ', 'saliency')),
        #     # Only positive if task=CLASS and has context
        #     TAG_IF('positive', ALL(
        #         CONDITION('TASK_EQ', 'class'),
        #         CONDITION('ALSO_HAS', [
        #             'Site Preparation', 'Active Construction', 'No Activity'],
        #         )
        #     )),
        # ],
    },
    {
        'name': 'No Activity',
        'color': 'tomato',
        'scored': True,
        'tags': ['saliency'],
        # 'conditional_tags': [
        #     TAG_IF('background', CONDITION('TASK_EQ', 'saliency')),
        #     # Only positive if task=CLASS and has context
        #     TAG_IF('positive', ALL(
        #         CONDITION('TASK_EQ', 'class'),
        #         CONDITION('ALSO_HAS', [
        #             'Site Preparation', 'Active Construction', 'No Activity'],
        #         )
        #     )),
        # ],
    },
]


def hack_track_categories(track_catnames, task):
    """
    Returns:
        List[str]: Modified categories

    Example:
        >>> from watch.heuristics import *  # NOQA
        >>> basis = {
        >>>     #'task': ['class', 'saliency'],
        >>>     'task': ['class'],
        >>>     'track_catnames': [
        >>>         ['No Activity'],
        >>>         ['Post Construction'],
        >>>         ['Post Construction', 'Post Construction', ],
        >>>         ['No Activity', 'ignore', 'ignore'],
        >>>         ['No Activity', 'Post Construction'],
        >>>         ['No Activity', 'Site Preparation', 'Post Construction'],
        >>>     ],
        >>> }
        >>> for kw in ub.named_product(basis):
        >>>     task = kw['task']
        >>>     track_catnames = kw['track_catnames']
        >>>     print('kw = {}'.format(ub.urepr(kw, nl=1)))
        >>>     print(hack_track_categories(track_catnames, task))

    Example:
        >>> from watch.heuristics import *  # NOQA
        >>> track_catnames = ['negative', 'negative']
        >>> task = 'saliency'
        >>> result = hack_track_categories(track_catnames, task)
        >>> print(result)
        ['negative', 'negative']

    """
    # FIXME! This is hard coded nonsense, need to come up with a general
    # way to encode these conditions in the categories themselves. Getting
    # this right is harder than I want it to be, so I'm hacking it.

    # Might want to make a real parser for this mini-language, or find an
    # existing mini-language that works

    main_classes = {'No Activity', 'Site Preparation', 'Active Construction', 'Post Construction'}

    # This is some of the uggliest code I've ever written
    new_catnames = []
    if task == 'saliency':
        for catname in track_catnames:
            if catname == 'No Activity':
                catname = 'background'
            elif catname == 'Post Construction':
                catname = 'background'
            elif catname == 'Unknown':
                catname = 'ignore'
            new_catnames.append(catname)
    elif task == 'class':
        unique_catnames = set(track_catnames)
        for catname in track_catnames:
            if catname == 'No Activity':
                remain = unique_catnames - {catname, None}
                if len(remain & main_classes) > 0:
                    catname = catname
                elif len(remain) == 0:
                    catname = 'background'
                elif remain.issubset({'background', 'ignore', 'Unknown'}):
                    catname = 'ignore'
            elif catname == 'Post Construction':
                remain = unique_catnames - {catname, None}
                if len(remain & main_classes) > 0:
                    catname = catname
                elif len(remain) == 0:
                    catname = 'background'
                elif remain.issubset({'background', 'ignore'}):
                    catname = 'ignore'
            elif catname == 'positive':
                catname = 'ignore'
            elif catname == 'negative':
                catname = 'background'
            elif catname == 'Unknown':
                catname = 'ignore'
            new_catnames.append(catname)
    else:
        raise KeyError(task)
        # op = condition['op']
        # # TODO: normalize classes
        # # TODO: make label conditionals as part of kwcoco
        # if op == 'ALSO_HAS':
        #     track_catnames = set(track_catnames)
        #     flag = any(arg in track_catnames for arg in condition['args'])
        #     return flag
        # else:
        #     raise NotImplementedError(op)
    return new_catnames


# Backwards compat (remove if nothing uses them)
CATEGORIES_SCORED = [c for c in CATEGORIES if c.get('scored', False)]
CATEGORIES_UNSCORED = [c for c in CATEGORIES if not c.get('scored', False)]

# CATEGORIES_UNSCORED = [
#     {'name': 'positive', 'color': 'olive', 'scored': False},
# ]
# CATEGORIES = CATEGORIES_SCORED + CATEGORIES_UNSCORED


# Might need to split this up into a finer-grained structure
# IGNORE_CLASSNAMES = {'clouds', 'occluded', 'ignore', 'unknown', 'Unknown'}
# BACKGROUND_CLASSES = {c['name'] for c in CATEGORIES if 'background' in c.get('tags', {})}
# UNDISTINGUISHED_CLASSES =  {c['name'] for c in CATEGORIES if 'saliency' in c.get('tags', {})}
IGNORE_CLASSNAMES = {'ignore', 'Unknown'}
BACKGROUND_CLASSES = {'background'}
NEGATIVE_CLASSES = {'negative'}
UNDISTINGUISHED_CLASSES =  {'positive'}
CONTEXT_CLASSES = {'No Activity', 'Post Construction'}
# 'background',
# 'No Activity',
# 'Post Construction',
# 'negative',
# }
# SPECIAL_CONTEXT_CLASSES = {
#     'No Activity',
#     'Post Construction',
#     'negative',
# }


# # These classes are used in BAS, but not in AC/SC
# UNDISTINGUISHED_CLASSES = {
#     'positive',
# }


CATEGORIES_DCT = {
    'positive': {
        'scored': [
            {'name': 'Site Preparation', 'color': 'gold'},
            {'name': 'Active Construction', 'color': 'lime'},
            {'name': 'Post Construction', 'color': 'darkturquoise'},

        ],
        'unscored': [
            {'name': 'positive', 'color': 'green'},
        ],
    },
    'negative': {
        'scored': [
            # Maybe this should not be marked as "scored", because it isn't
            {'name': 'No Activity', 'color': 'tomato'},
        ],
        'unscored': [
            {'name': 'Unknown', 'color': 'blueviolet'},
            {'name': 'ignore', 'color': 'slategray'},
            {'name': 'negative', 'color': 'orangered'},
            {'name': 'background', 'color': 'black'},
        ],
    }
}

# 'name' field only
CNAMES_DCT = {
    k1: {k2: [cat['name'] for cat in cats]
         for k2, cats in dct.items()}
    for k1, dct in CATEGORIES_DCT.items()
}


# For passing site summaries from BAS to SC
SITE_SUMMARY_CNAME = 'Site Boundary'


def ensure_heuristic_coco_colors(coco_dset, force=False):
    """
    Args:
        coco_dset (kwcoco.CocoDataset): object to modify
        force (bool): if True, overwrites existing colors if needed

    TODO:
        - [ ] Move this non-heuristic functionality to
            :func:`kwcoco.CocoDataset.ensure_class_colors`

    Example:
        >>> from watch.heuristics import *  # NOQA
        >>> import kwcoco
        >>> coco_dset = kwcoco.CocoDataset.demo()
        >>> ensure_heuristic_coco_colors(coco_dset)
        >>> assert all(c['color'] for c in coco_dset.cats.values())
    """
    for hcat in CATEGORIES:
        cat = coco_dset.index.name_to_cat.get(hcat['name'], None)
        if cat is not None:
            if force or cat.get('color', None) is None:
                cat['color'] = hcat['color']
    data_dicts = coco_dset.dataset['categories']
    _ensure_distinct_dict_colors(data_dicts)


def ensure_heuristic_category_tree_colors(classes, force=False):
    """
    Args:
        classes (kwcoco.CategoryTree): object to modify
        force (bool): if True, overwrites existing colors if needed

    TODO:
        - [ ] Move this non-heuristic functionality to
            :func:`kwcoco.CategoryTree.ensure_colors`

    Example:
        >>> from watch.heuristics import *  # NOQA
        >>> import kwcoco
        >>> classes = kwcoco.CategoryTree.coerce(['ignore', 'positive', 'Active Construction', 'foobar', 'Unknown', 'baz'])
        >>> ensure_heuristic_category_tree_colors(classes)
        >>> assert all(d['color'] for n, d in classes.graph.nodes(data=True))
    """
    # Set any missing class color with the heuristic category
    for hcat in CATEGORIES:
        node_data = classes.graph.nodes.get(hcat['name'], None)
        if node_data is not None:
            if force or node_data.get('color', None) is None:
                node_data['color'] = hcat['color']
    data_dicts = [data for node, data in classes.graph.nodes(data=True)]
    _ensure_distinct_dict_colors(data_dicts)


def _ensure_distinct_dict_colors(data_dicts, force=False):
    # Generalized part that could move to kwcoco
    have_dicts = [d for d in data_dicts if d.get('color', None) is not None]
    miss_dicts = [d for d in data_dicts if d.get('color', None) is None]
    num_uncolored = len(miss_dicts)
    if num_uncolored:
        import kwimage
        existing_colors = [kwimage.Color(d['color']).as01() for d in have_dicts]
        new_colors = kwimage.Color.distinct(
            num_uncolored, existing=existing_colors, legacy=False)
        for d, c in zip(miss_dicts, new_colors):
            d['color'] = c


HUERISTIC_COMBINABLE_CHANNELS = [
    ub.oset(['B04', 'B03', 'B02']),  # for onera
    ub.oset(['matseg_1', 'matseg_2', 'matseg_3']),  # hack
    # ub.oset(['snow_or_ice_field', 'built_up', 'grassland']),  # hack
]

CONFUSION_COLOR_SCHEME = {
    'TN': 'black',
    # 'TP': 'white',
    # 'TP': 'snow',  # off white
    'TP': 'whitesmoke',  # off white
    'FN': 'teal',
    'FP': 'red',
}


def dummy_legend():
    # hack to make a legend for slides
    """
    from watch.heuristics import *  # NOQA
    dummy_legend()

    """
    import kwplot
    kwplot.autompl()
    img = kwplot.make_legend_img(CONFUSION_COLOR_SCHEME)
    kwplot.imshow(img, fnum=1)

    label_to_color = {
        cat['name']: cat['color']
        for cat in CATEGORIES
    }
    img = kwplot.make_legend_img(label_to_color)
    kwplot.imshow(img, fnum=2)

    label_to_color = ub.dict_subset(label_to_color, {
        'Post Construction', 'Site Preparation', 'Active Construction',
        'No Activity', 'Unknown'})
    img = kwplot.make_legend_img(label_to_color)
    kwplot.imshow(img, fnum=3)


def build_image_header_text(**kwargs):
    """
    A heuristic for what sort of info is useful to plot on the header of an
    image.

    Kwargs:
        img
        coco_dset
        vidname,
        _header_extra

        gid,
        frame_index,
        dset_idstr,
        name,
        sensor_coarse,
        date_captured

    Example:
        >>> from watch.heuristics import *  # NOQA
        >>> img = {
        >>>     'id': 1,
        >>>     'frame_index': 0,
        >>>     'date_captured': '2020-01-01',
        >>>     'name': 'BLARG',
        >>>     'sensor_coarse': 'Sensor1',
        >>> }
        >>> kwargs = {
        >>>     'img': img,
        >>>     'dset_idstr': '',
        >>>     'name': '',
        >>>     '_header_extra': None,
        >>> }
        >>> header_lines = build_image_header_text(**kwargs)
        >>> print('header_lines = {}'.format(ub.urepr(header_lines, nl=1)))
    """
    img = kwargs.get('img', {})
    _header_extra = kwargs.get('_header_extra', None)
    dset_idstr = kwargs.get('dset_idstr', '')

    def _multi_get(key, default=ub.NoParam, *dicts):
        # try to lookup from multiple dictionaries
        found = default
        for d in dicts:
            if key in d:
                found = d[key]
                break
        if found is ub.NoParam:
            raise Exception
        return found

    sensor_coarse = _multi_get('sensor_coarse', 'unknown', kwargs, img)
    # name = _multi_get('name', 'unknown', kwargs, img)

    date_captured = _multi_get('date_captured', '', kwargs, img)
    frame_index = _multi_get('frame_index', None, kwargs, img)
    gid = _multi_get('id', None, kwargs, img)
    image_name = _multi_get('name', '', kwargs, img)

    vidname = None
    if 'vidname' in kwargs:
        vidname = kwargs['vidname']
    else:
        coco_dset = kwargs.get('coco_dset', None)
        if coco_dset is not None:
            vidname = coco_dset.index.videos[img['video_id']]['name']

    image_id_parts = []
    image_id_parts.append(f'gid={gid}')
    image_id_parts.append(f'frame_index={frame_index}')
    image_id_part = ', '.join(image_id_parts)

    header_line_infos = []
    header_line_infos.append([vidname, image_id_part, _header_extra])
    header_line_infos.append([dset_idstr])
    header_line_infos.append([image_name])
    header_line_infos.append([sensor_coarse, date_captured])
    header_lines = []
    for line_info in header_line_infos:
        header_line = ' '.join([p for p in line_info if p])
        header_line = header_line.replace('\\n', '\n')  # hack
        if header_line:
            header_lines.append(header_line)
    return header_lines


# NOTES ON QUALITY / CLOUDMASK
# https://github.com/GERSL/Fmask#46-version
# The cloudmask band is a class-idx based raster with labels
# 0 => clear land pixel
# 1 => clear water pixel
# 2 => cloud shadow
# 3 => snow
# 4 => cloud
# 255 => no observation

# However, in my data I seem to see:
# Unique values   8,  16,  65, 128

# These are specs
# https://smartgitlab.com/TE/standards/-/wikis/Data-Output-Specifications#quality-band
# TODO: this could be a specially handled frame like ASI.
# QA Information Moved to ~/code/watch/watch/tasks/fusion/datamodules/qa_bands.py
QUALITY_BITS = ub.udict({
    'TnE'           : 1 << 0,  # T&E binary mask
    'dilated_cloud' : 1 << 1,
    'cirrus'        : 1 << 2,
    'cloud'         : 1 << 3,
    'cloud_shadow'  : 1 << 4,
    'snow'          : 1 << 5,
    'clear'         : 1 << 6,
    'water'         : 1 << 7,
})


# The main dataset codes currently in use.
DATASET_CODES = [
    # 'Aligned-Drop4-2022-08-08-TA1-S2-L8-ACC',
    # 'Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC',
    'Drop4-BAS',
    'Drop4-SC',
]


DVC_FIND_EXPT_KWARGS = {
    'tags': 'phase2_expt', 'envvar': 'DVC_EXPT_DPATH', 'hardware': 'auto'}
DVC_FIND_DATA_KWARGS = {
    'tags': 'phase2_data', 'envvar': 'DVC_DATA_DPATH', 'hardware': 'auto'}


def auto_expt_dvc():
    import watch
    return watch.find_smart_dvc_dpath(**DVC_FIND_EXPT_KWARGS)


def auto_data_dvc():
    import watch
    return watch.find_smart_dvc_dpath(**DVC_FIND_DATA_KWARGS)


# We should be able to figure out a way to robustly introspect these
fit_param_keys = [
    'sensorchan',
    # 'channels',
    'time_steps',
    'chip_dims', 'chip_overlap', 'arch_name', 'optimizer',
    'time_sampling', 'time_span', 'true_multimodal',
    'accumulate_grad_batches', 'modulate_class_weights', 'tokenizer',
    'use_grid_positives', 'use_cloudmask', 'upweight_centers',
    'temporal_dropout', 'stream_channels', 'saliency_loss',
    'class_loss', 'init', 'learning_rate', 'decoder',
]


pred_param_keys = [
    'pred_tta_fliprot',
    'pred_tta_time',
    'pred_chip_overlap',
]


trk_param_keys = [
    'trk_thresh',
    'trk_morph_kernel',
    'trk_agg_fn',
    'trk_thresh_hysteresis',
    'trk_moving_window_size',
]


act_param_keys = [
    'trk_use_viterbi',
    'trk_thresh',
]


DSET_CODE_TO_GSD = {
    # DEPRECATE
    'Aligned-Drop3-L1': 10.0,
    'Aligned-Drop3-TA1-2022-03-10': 10.0,
    'Cropped-Drop3-TA1-2022-03-10': 1.0,
}


PHASES = ['No Activity', 'Site Preparation', 'Active Construction', 'Post Construction']


# Represents default relative weights to give to each sensor in temporal
# sampling.
SENSOR_TEMPORAL_SAMPLING_VALUES = {
    'WV': 10,
    'WV1': 9,
    'S2': 1,
    'PD': 7,
    'L8': 0.3,
    'sensor1': 11,
    'sensor2': 7,
    'sensor3': 5,
    'sensor4': 3,
}


# Hard coded values for which regions are cleared
REGION_STATUS = [
    {'region_id': 'AE_R001', 'cleared': True},
    {'region_id': 'BH_R001', 'cleared': True},
    {'region_id': 'BR_R001', 'cleared': True},
    {'region_id': 'BR_R002', 'cleared': True},
    {'region_id': 'BR_R004', 'cleared': True},
    {'region_id': 'BR_R005', 'cleared': True},
    {'region_id': 'CH_R001', 'cleared': True},
    {'region_id': 'KR_R001', 'cleared': True},
    {'region_id': 'KR_R002', 'cleared': True},
    {'region_id': 'LT_R001', 'cleared': True},
    {'region_id': 'NZ_R001', 'cleared': True},
    {'region_id': 'PE_R001', 'cleared': True},
    {'region_id': 'US_R001', 'cleared': True},
    {'region_id': 'US_R004', 'cleared': True},
    {'region_id': 'US_R005', 'cleared': True},
]


# Mapping from our sensor names to the official T&E sensor names
TE_SENSOR_NAMES = {
    'WV': 'WorldView',
    'S2': 'Sentinel-2',
    'LE': 'Landsat 7',
    'LC': 'Landsat 8',
    'L8': 'Landsat 8',
    'WV1': 'WorldView 1',
    'PD': 'Planet',
}



import argparse
import sys
import json
import os
import tempfile
import subprocess


def main():
    parser = argparse.ArgumentParser(
        description="Egress KWCOCO data to T&E baseline framework structure")

    parser.add_argument('input_path',
                        type=str,
                        help="Path to input T&E Baseline Framework JSON")
    parser.add_argument("-o", "--outdir",
                        type=str,
                        required=True,
                        help="Output directory for ingressed assets an output "
                             "STAC Catalog")
    parser.add_argument("--aws_profile",
                        required=False,
                        type=str,
                        help="AWS Profile to use for AWS S3 CLI commands")
    parser.add_argument("-d", "--dryrun",
                        action='store_true',
                        default=False,
                        help="Run AWS CLI commands with --dryrun flag")
    parser.add_argument('-s', '--show-progress',
                        action='store_true',
                        default=False,
                        help='Show progress for AWS CLI commands')

    baseline_framework_kwcoco_ingress(**vars(parser.parse_args()))

    return 0


def baseline_framework_kwcoco_ingress(input_path,
                                      outdir,
                                      aws_profile=None,
                                      dryrun=False,
                                      show_progress=False):
    os.makedirs(outdir, exist_ok=True)

    if aws_profile is not None:
        aws_base_command =\
            ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    if dryrun:
        aws_base_command.append('--dryrun')

    if not show_progress:
        aws_base_command.append('--only-show-errors')

    def _load_input(path):
        try:
            with open(path) as f:
                input_json = json.load(f)
            return input_json['stac'].get('features', [])
        # Excepting KeyError here in case of a single line STAC item input
        except (json.decoder.JSONDecodeError, KeyError):
            # Support for simple newline separated STAC items
            with open(path) as f:
                return [json.loads(line) for line in f]

    if input_path.startswith('s3'):
        with tempfile.NamedTemporaryFile() as temporary_file:
            subprocess.run(
                [*aws_base_command, input_path, temporary_file.name],
                check=True)

            input_stac_items = _load_input(temporary_file.name)
    else:
        input_stac_items = _load_input(input_path)

    # Our baseline KWCOCO egress script should only ever write out a
    # single KWCOCO STAC item
    if len(input_stac_items) != 1:
        raise RuntimeError("Expecting one and only one STAC item from input")
    kwcoco_stac_item = input_stac_items[0]

    try:
        kwcoco_dataset_href = kwcoco_stac_item['assets']['kwcoco']['href']
    except KeyError:
        raise RuntimeError("Expecting asset named 'kwcoco' in input "
                           "KWCOCO STAC item")

    # Assumes that all the necessary KWCOCO dataset assets are in the
    # same directory as the KWCOCO dataset itself
    kwcoco_dataset_dir = os.path.dirname(kwcoco_dataset_href)
    subprocess.run([*aws_base_command, '--recursive',
                    kwcoco_dataset_dir, outdir],
                   check=True)

    # Returns local path to retreived KWCOCO dataset
    return os.path.join(outdir, os.path.basename(kwcoco_dataset_href))


if __name__ == "__main__":
    sys.exit(main())



from dataclasses import dataclass
from watch.tasks.tracking.utils import TrackFunction


class MonoTrack(TrackFunction):
    '''
    Combine all polygons into the same track.
    '''

    def __init__(self, **kwargs):
        self.kwargs = kwargs  # Unused

    def __call__(self, coco_dset):
        from watch.utils.kwcoco_extensions import TrackidGenerator
        coco_dset.annots().set('track_id', next(TrackidGenerator(coco_dset)))

        return coco_dset


def as_shapely_polys(annots):
    return map(lambda poly: poly.to_shapely().buffer(0),
               annots.detections.data['segmentations'].to_polygon_list())


@dataclass
class OverlapTrack(TrackFunction):
    '''
    Put polygons in the same track if their areas overlap.
    '''
    min_overlap: float = 0

    def __call__(self, coco_dset):
        from watch.utils.kwcoco_extensions import TrackidGenerator
        new_trackids = TrackidGenerator(coco_dset)

        annots = coco_dset.annots()

        aid_to_poly = dict(zip(annots.aids, as_shapely_polys(annots)))

        def _search(aid, aid_groups):
            poly1 = aid_to_poly[aid]

            def _search_group(aids):
                for aid2 in aids:
                    if 'track_id' not in coco_dset.anns[aid2]:
                        poly2 = aid_to_poly[aid2]
                        # check overlap
                        if poly1.intersects(poly2):
                            if (poly1.intersection(poly2).area /
                                    poly2.area) > self.min_overlap:
                                return aid2

            try:
                return next(filter(None, map(_search_group, aid_groups)))
            except StopIteration:
                return None

        # update tracks one frame at a time
        aids_by_frame = list(
            map(coco_dset.gid_to_aids.get,
                coco_dset.index._set_sorted_by_frame_index(coco_dset.imgs)))

        for frame_ix, aids in enumerate(aids_by_frame):

            for aid in aids:

                ann = coco_dset.anns[aid]
                if 'track_id' not in ann:
                    trackid = next(new_trackids)
                    ann['track_id'] = trackid
                else:
                    trackid = ann['track_id']

                next_aid = _search(aid, aids_by_frame[frame_ix + 1:])
                if next_aid is not None:
                    next_ann = coco_dset.anns[next_aid]
                    next_ann['track_id'] = trackid

                DEBUG_JSON_SERIALIZABLE = 0
                if DEBUG_JSON_SERIALIZABLE:
                    from kwcoco.util import util_json
                    unserializable = list(util_json.find_json_unserializable(next_ann))
                    if unserializable:
                        import ubelt as ub
                        raise Exception('Inside OverlapTrack: ' + ub.urepr(unserializable))

        return coco_dset



import torch.nn as nn


class SimpleCNNEncoder(nn.Module):
    def __init__(self, in_channels, bilinear=True):
        super(SimpleCNNEncoder, self).__init__()
        self.bilinear = bilinear
        self.in_channels = in_channels

        self.build()

    def build(self):
        self.inc = DoubleConv(self.in_channels, 64)
        self.down1 = Down(64, 128)
        self.down2 = Down(128, 256)
        self.down3 = Down(256, 512)
        factor = 2 if self.bilinear else 1
        self.down4 = Down(512, 1024 // factor)

    def forward(self, input):
        output = {}

        enc_input = self.inc(input)
        output["layer1"] = self.down1(enc_input)
        output["layer2"] = self.down2(output["layer1"])
        output["layer3"] = self.down3(output["layer2"])
        output["layer4"] = self.down4(output["layer3"])

        return output


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(nn.MaxPool2d(2), DoubleConv(in_channels, out_channels))

    def forward(self, x):
        return self.maxpool_conv(x)
